# -*- coding: utf-8 -*-
"""Base Site class."""
from typing import ClassVar
from typing import Dict
from typing import List

from mrsimulator.utils.parseable import Parseable
from pydantic import validator

from .tensors import AntisymmetricTensor
from .tensors import SymmetricTensor

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"


class Coupling(Parseable):
    """Base class representing a two-site coupled nuclear spin interaction tensor
    parameters, which include the J-coupling and dipolar tensor.

    .. rubric:: Attribute Documentation

    Attributes
    ----------

    site_index: list of int (required).
        A list of two integers, each corresponding to the index of the coupled sites.

        Example
        -------

        >>> coupling = Coupling(site_index=[0, 1])

    isotropic_j: float (optional).
        The isotropic j-coupling, in Hz, between the coupled sites. The default is 0.

        Example
        -------

        >>> coupling.isotropic_j = 43.3

    j_symmetric: :ref:`sy_api` or equivalent dict object (optional).
        The attribute represents the parameters of the irreducible second-rank traceless
        symmetric part of the J-coupling tensor. The default value is None.

        The allowed attributes of the :ref:`sy_api` class for `j_symmetric` are
        ``zeta``, ``eta``, ``alpha``, ``beta``, and ``gamma``, where ``zeta`` is the
        J anisotropy, in Hz, and ``eta`` is the J asymmetry parameter defined using the
        Haeberlen convention. The Euler angles ``alpha``, ``beta``, and ``gamma`` are
        in radians.

        Example
        -------

        >>> coupling.j_symmetric = {'zeta': 10, 'eta': 0.5}

        >>> # or equivalently
        >>> coupling.j_symmetric = SymmetricTensor(zeta=10, eta=0.5)

    j_antisymmetric: :ref:`asy_api` or equivalent dict object (optional).
        The attribute represents the parameters of the irreducible first-rank
        antisymmetric part of the J tensor. The default value is None.

        The allowed attributes of the :ref:`asy_api` class for `j_antisymmetric` are
        ``zeta``, ``alpha``, and ``beta``, where ``zeta`` is the anisotropy parameter
        of the anti-symmetric first-rank tensor given in Hz. The angles ``alpha`` and
        ``beta`` are in radians.

        Example
        -------

        >>> coupling.j_antisymmetric = {'zeta': 20}

        >>> # or equivalently
        >>> coupling.j_antisymmetric = AntisymmetricTensor(zeta=20)

    dipolar: :ref:`sy_api` or equivalent dict object (optional).
        The attribute represents the parameters of the irreducible second-rank traceless
        symmetric part of the direct-dipolar coupling tensor. The default value is None.

        The allowed attributes of the :ref:`sy_api` class for `dipolar` are ``D``,
        ``alpha``, ``beta``, and ``gamma``, where ``D`` is the dipolar coupling
        constant, in Hz. The Euler angles ``alpha``, ``beta``, and ``gamma`` are in
        radians.

        Example
        -------

        >>> coupling.dipolar = {'D': 320}

        >>> # or equivalently
        >>> coupling.dipolar = SymmetricTensor(D=320)

    name: str (optional).
        The name or id of the coupling. The default value is None.

        Example
        -------

        >>> coupling.name = '1H-1H'
        >>> coupling.name
        '1H-1H'

    label: str (optional).
        The label for the coupling. The default value is None.

        Example
        -------

        >>> coupling.label = 'Weak coupling'
        >>> coupling.label
        'Weak coupling'

    description: str (optional).
        A description of the coupling. The default value is None.

        Example
        -------

        >>> coupling.description = 'An example coupled sites.'
        >>> coupling.description
        'An example coupled sites.'

    Example
    -------

    The following are a few examples of setting the site object.

    >>> coupling1 = Coupling(
    ...     site_index=[0, 1],
    ...     isotropic_j=20, # in Hz
    ...     j_symmetric={
    ...         "zeta": 10, # in Hz
    ...         "eta": 0.5
    ...     },
    ...     dipolar={"D": 5.1e3}, # in Hz
    ... )

    Using SymmetricTensor objects.

    >>> coupling1 = Coupling(
    ...     site_index=[0, 1],
    ...     isotropic_j=20, # in Hz
    ...     j_symmetric=SymmetricTensor(zeta=10, eta=0.5),
    ...     dipolar=SymmetricTensor(D=5.1e3), # in Hz
    ... )
    """

    name: str = None
    label: str = None
    description: str = None
    site_index: List[int]
    isotropic_j: float = 0.0
    j_symmetric: SymmetricTensor = None
    j_antisymmetric: AntisymmetricTensor = None
    dipolar: SymmetricTensor = None

    property_unit_types: ClassVar = {"isotropic_j": "frequency"}
    property_default_units: ClassVar = {"isotropic_j": "Hz"}
    property_units: Dict = {"isotropic_j": "Hz"}

    @validator("dipolar")
    def dipolar_must_not_contain_Cq_and_zeta(cls, v, values):
        if v is None:
            return v
        _ = [
            v.property_units.pop(item) if item in v.property_units else None
            for item in ["Cq", "zeta"]
        ]
        return v

    @validator("j_symmetric", "j_antisymmetric")
    def j_symmetric_must_not_contain_Cq_and_D(cls, v, values):
        if v is None:
            return v
        _ = [
            v.property_units.pop(item) if item in v.property_units else None
            for item in ["Cq", "D"]
        ]
        v.property_units["zeta"] = "Hz"
        return v

    @validator("site_index", always=True)
    def validate_site_index(cls, v, *, values, **kwargs):
        if len(v) != 2:
            raise ValueError("Site index must a list of two integers.")
        if v[0] == v[1]:
            raise ValueError("The two site indexes must be unique integers.")
        return v

    class Config:
        validate_assignment = True

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """Parse the physical quantity from a dictionary representation of the Coupling
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict py_dict: A required python dict object.

        Returns:
            :ref:`site_api` object.

        Example
        -------

        >>> coupling_dict = {
        ...    "site_index": [1, 2],
        ...    "isotropic_j": "20 Hz",
        ...    "j_symmetric": {"zeta": "10 Hz", "eta":0.5}
        ... }
        >>> coupling1 = Coupling.parse_dict_with_units(coupling_dict)
        """
        prop_mapping = {
            "j_symmetric": SymmetricTensor,
            "j_antisymmetric": AntisymmetricTensor,
            "dipolar": SymmetricTensor,
        }

        for k, v in prop_mapping.items():
            if k in py_dict:
                py_dict[k] = v.parse_dict_with_units(py_dict[k])
                if py_dict[k].property_units["zeta"] == "ppm" and k != "dipolar":
                    raise ValueError(
                        f"Error enforcing units for {k}.zeta: ppm. Use frequency units."
                    )
        return super().parse_dict_with_units(py_dict)



# -*- coding: utf-8 -*-
"""Base Simulator class."""
import json
from copy import deepcopy
from typing import List

import csdmpy as cp
import numpy as np
import pandas as pd
import psutil
from joblib import delayed
from joblib import Parallel
from mrsimulator import __version__
from mrsimulator import methods as NamedMethods
from mrsimulator import Site
from mrsimulator import SpinSystem
from mrsimulator.base_model import one_d_spectrum
from mrsimulator.method import Method
from mrsimulator.spin_system.isotope import Isotope
from mrsimulator.utils import flatten_dict
from mrsimulator.utils.abstract_list import AbstractList
from mrsimulator.utils.extra import _reduce_dict
from mrsimulator.utils.importer import import_json
from pydantic import BaseModel

from .config import ConfigSimulator

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"

__CPU_count__ = psutil.cpu_count()

__named_methods__ = [
    val for k, val in NamedMethods.__dict__.items() if isinstance(val, type)
]
__method_names__ = [item.__name__ for item in __named_methods__]
__sim_methods__ = {k: v for k, v in zip(__method_names__, __named_methods__)}


class Simulator(BaseModel):
    """The simulator class.

    Attributes
    ----------

    spin_systems: A list of :ref:`spin_sys_api` or equivalent dict objects (optional).
        A list of :ref:`spin_sys_api` or equivalent dict objects representing a
        collection of isolated NMR spin systems present within the sample. The default
        value is an empty list.

        Example
        -------

        >>> sim = Simulator()
        >>> sim.spin_systems = [
        ...     SpinSystem(sites=[Site(isotope='17O')], abundance=0.015),
        ...     SpinSystem(sites=[Site(isotope='1H')], abundance=1),
        ... ]

        >>> # or equivalently
        >>> sim.spin_systems = [
        ...     {'sites': [{'isotope': '17O'}], 'abundance': 0.015},
        ...     {'sites': [{'isotope': '1H'}], 'abundance': 1},
        ... ]

    methods: A list of :ref:`method_api` or equivalent dict objects (optional).
        A list of :ref:`method_api`  or equivalent dict objects representing an NMR
        methods. The default value is an empty list.

        Example
        -------

        >>> from mrsimulator.methods import BlochDecaySpectrum
        >>> from mrsimulator.methods import BlochDecayCTSpectrum
        >>> sim.methods = [
        ...     BlochDecaySpectrum(channels=['17O'], spectral_width=50000),
        ...     BlochDecayCTSpectrum(channels=['17O'])
        ... ]

    config: :ref:`config_api` object or equivalent dict object (optional).
        The :ref:`config_api` object is used to configure the simulation. The valid
        attributes of the ConfigSimulator object are

        - ``number_of_sidebands``,
        - ``integration_density``,
        - ``integration_volume``, and
        - ``decompose_spectrum``

        Example
        -------

        >>> from mrsimulator.simulator.config import ConfigSimulator
        >>> sim.config = ConfigSimulator(
        ...     number_of_sidebands=32,
        ...     integration_density=64,
        ...     integration_volume='hemisphere',
        ...     decompose_spectrum='spin_system',
        ... )

        >>> # or equivalently
        >>> sim.config = {
        ...     'number_of_sidebands': 32,
        ...     'integration_density': 64,
        ...     'integration_volume': 'hemisphere',
        ...     'decompose_spectrum': 'spin_system',
        ... }

        See :ref:`config_simulator` for details.

    name: str (optional).
        The name or id of the simulation or sample. The default value is None.

        Example
        -------

        >>> sim.name = '1H-17O'
        >>> sim.name
        '1H-17O'

    label: str (optional).
        The label for the simulation or sample. The default value is None.

        Example
        -------

        >>> sim.label = 'Test simulator'
        >>> sim.label
        'Test simulator'

    description: str (optional).
        A description of the simulation or sample. The default value is None.

        Example
        -------

        >>> sim.description = 'Simulation for sample 1'
        >>> sim.description
        'Simulation for sample 1'

    """

    name: str = None
    label: str = None
    description: str = None
    spin_systems: List[SpinSystem] = []
    methods: List[Method] = []
    config: ConfigSimulator = ConfigSimulator()
    indexes = []

    class Config:
        validate_assignment = True

    def __eq__(self, other):
        if not isinstance(other, Simulator):
            return False
        check = [
            self.name == other.name,
            self.label == other.label,
            self.description == other.description,
            self.spin_systems == other.spin_systems,
            self.methods == other.methods,
            self.config == other.config,
            np.all(self.indexes == other.indexes),
        ]
        if np.all(check):
            return True
        return False

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """Parse the physical quantity from a dictionary representation of the Simulator
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict py_dict: A required python dict object.

        Returns:
            A :ref:`simulator_api` object.

        Example
        -------

        >>> sim_py_dict = {
        ...     'config': {
        ...         'decompose_spectrum': 'none',
        ...         'integration_density': 70,
        ...         'integration_volume': 'octant',
        ...         'number_of_sidebands': 64
        ...     },
        ...     'spin_systems': [
        ...         {
        ...             'abundance': '100 %',
        ...             'sites': [{
        ...                 'isotope': '13C',
        ...                 'isotropic_chemical_shift': '20.0 ppm',
        ...                 'shielding_symmetric': {'eta': 0.5, 'zeta': '10.0 ppm'}
        ...             }]
        ...         },
        ...         {
        ...             'abundance': '100 %',
        ...             'sites': [{
        ...                 'isotope': '1H',
        ...                     'isotropic_chemical_shift': '-4.0 ppm',
        ...                     'shielding_symmetric': {'eta': 0.1, 'zeta': '2.1 ppm'}
        ...             }]
        ...         },
        ...         {
        ...             'abundance': '100 %',
        ...             'sites': [{
        ...                 'isotope': '27Al',
        ...                 'isotropic_chemical_shift': '120.0 ppm',
        ...                 'shielding_symmetric': {'eta': 0.1, 'zeta': '2.1 ppm'}
        ...             }]
        ...         }
        ...     ]
        ... }
        >>> sim = Simulator.parse_dict_with_units(sim_py_dict)
        >>> len(sim.spin_systems)
        3
        """
        py_copy_dict = deepcopy(py_dict)

        if "spin_systems" in py_copy_dict:
            spin_sys = py_copy_dict["spin_systems"]
            spin_sys = [SpinSystem.parse_dict_with_units(obj) for obj in spin_sys]
            py_copy_dict["spin_systems"] = spin_sys

        if "methods" in py_copy_dict:
            methods = py_copy_dict["methods"]
            method_cls = [
                Method
                if obj["name"] not in __method_names__
                else __sim_methods__[obj["name"]]
                for obj in methods
            ]

            methods = [
                fn.parse_dict_with_units(obj) for obj, fn in zip(methods, method_cls)
            ]
            py_copy_dict["methods"] = methods

        return Simulator(**py_copy_dict)

    def get_isotopes(self, spin_I: float = None, symbol: bool = False) -> list:
        """List of unique isotopes from the sites within the list of the spin systems
        corresponding to spin quantum number `I`. If `I` is None, a list of all unique
        isotopes is returned instead.

        Args:
            float spin_I: An optional spin quantum number. The valid input are the
                multiples of 0.5.
            bool symbol: If true, return a list of str with isotope symbols.

        Returns:
            A list of :ref:`isotope_api` objects.

        Example
        -------

        >>> sim.get_isotopes()
        [Isotope(symbol='13C'), Isotope(symbol='1H'), Isotope(symbol='27Al')]
        >>> sim.get_isotopes(symbol=True)
        ['13C', '1H', '27Al']

        >>> sim.get_isotopes(spin_I=0.5)
        [Isotope(symbol='13C'), Isotope(symbol='1H')]
        >>> sim.get_isotopes(spin_I=0.5, symbol=True)
        ['13C', '1H']

        >>> sim.get_isotopes(spin_I=1.5)
        []

        >>> sim.get_isotopes(spin_I=2.5)
        [Isotope(symbol='27Al')]
        >>> sim.get_isotopes(spin_I=2.5, symbol=True)
        ['27Al']
        """
        st = []
        for sys in self.spin_systems:
            st += sys.get_isotopes(spin_I, symbol=True)
        st = np.unique(st)
        if not symbol:
            return [Isotope(symbol=item) for item in st]
        return list(st)

    def json(self, include_methods: bool = False, include_version: bool = False):
        """Parse the class object to a JSON compliant python dictionary object, where
        the attribute value with physical quantity is expressed as a string with a
        value and a unit.

        Args:
            bool include_methods: If True, the output dictionary will include the
                serialized method objects. The default value is False.
            bool include_version: If True, add a version key-value pair to the
                serialized output dictionary. The default is False.

        Returns:
            A Dict object.

        Example
        -------

        >>> pprint(sim.json())
        {'config': {'decompose_spectrum': 'none',
                    'integration_density': 70,
                    'integration_volume': 'octant',
                    'number_of_sidebands': 64},
         'spin_systems': [{'abundance': '100.0 %',
                           'sites': [{'isotope': '13C',
                                      'isotropic_chemical_shift': '20.0 ppm',
                                      'shielding_symmetric': {'eta': 0.5,
                                                              'zeta': '10.0 ppm'}}]},
                          {'abundance': '100.0 %',
                           'sites': [{'isotope': '1H',
                                      'isotropic_chemical_shift': '-4.0 ppm',
                                      'shielding_symmetric': {'eta': 0.1,
                                                              'zeta': '2.1 ppm'}}]},
                          {'abundance': '100.0 %',
                           'sites': [{'isotope': '27Al',
                                      'isotropic_chemical_shift': '120.0 ppm',
                                      'shielding_symmetric': {'eta': 0.1,
                                                              'zeta': '2.1 ppm'}}]}]}
        """
        sim = {}
        sim["name"] = self.name
        sim["description"] = self.description
        sim["label"] = self.label
        sim["spin_systems"] = [_.json() for _ in self.spin_systems]
        sim["methods"] = [_.json() for _ in self.methods] if include_methods else None
        sim["config"] = self.config.dict()
        sim["version"] = __version__ if include_version else None

        _ = [sim.pop(k) for k in [k for k in sim.keys() if sim[k] is None]]
        return sim

    def reduced_dict(self, exclude=["property_units", "indexes"]) -> dict:
        """Returns a reduced dictionary representation of the class object by removing
        all key-value pair corresponding to keys listed in the `exclude` argument, and
        keys with value as None.

        Args:
            list exclude: A list of keys to exclude from the dictionary.
        Return: A dict.
        """
        return _reduce_dict(self.dict(), exclude)

    # def pretty(self):
    #     return JSON(self.json(include_methods=True, include_version=True))

    def load_spin_systems(self, filename: str):
        """Load a list of spin systems from the given JSON serialized file.

        See an
        `example <https://raw.githubusercontent.com/DeepanshS/mrsimulator-examples/
        master/spin_systems_v0.3.json>`_ of a JSON serialized file. For details, refer
        to the :ref:`load_spin_systems` section of this documentation.

        Args:
            str filename: A local or remote address to a JSON serialized file.

        Example
        -------

        >>> sim.load_spin_systems(filename) # doctest:+SKIP
        """
        contents = import_json(filename)
        # json_data = contents["spin_systems"]
        self.spin_systems = [SpinSystem.parse_dict_with_units(obj) for obj in contents]

    def export_spin_systems(self, filename: str):
        """Export a list of spin systems to a JSON serialized file.

        See an
        `example <https://raw.githubusercontent.com/DeepanshS/mrsimulator-examples/
        master/spin_systems_v0.3.json>`_ of a JSON serialized file. For details, refer
        to the :ref:`load_spin_systems` section.

        Args:
            str filename: A filename of the serialized file.

        Example
        -------

        >>> sim.export_spin_systems(filename) # doctest:+SKIP
        """
        spin_sys = [SpinSystem.json(obj) for obj in self.spin_systems]
        with open(filename, "w", encoding="utf8") as outfile:
            json.dump(
                spin_sys, outfile, ensure_ascii=False, sort_keys=False, allow_nan=False
            )

    def run(
        self,
        method_index: list = None,
        n_jobs: int = 1,
        pack_as_csdm: bool = True,
        **kwargs,
    ):
        """Run the simulation and compute spectrum.

        Args:
            method_index: An integer or a list of integers. If provided, only the
                simulations corresponding to the methods at the given index/indexes
                will be computed. The default is None, `i.e.`, the simulation for
                all method will be computed.
            bool pack_as_csdm: If true, the simulation results are stored as a
                `CSDM <https://csdmpy.readthedocs.io/en/stable/api/CSDM.html>`_ object,
                otherwise, as a `ndarray
                <https://numpy.org/doc/1.18/reference/generated/numpy.ndarray.html>`_
                object.
                The simulations are stored as the value of the
                :attr:`~mrsimulator.Method.simulation` attribute of the corresponding
                method.

        Example
        -------

        >>> sim.run() # doctest:+SKIP
        """
        verbose = 0
        if method_index is None:
            method_index = np.arange(len(self.methods))
        elif isinstance(method_index, int):
            method_index = [method_index]
        for index in method_index:
            method = self.methods[index]
            spin_sys = get_chunks(self.spin_systems, n_jobs)
            kwargs_dict = self.config.get_int_dict()
            jobs = (
                delayed(one_d_spectrum)(
                    method=method, spin_systems=sys, **kwargs_dict, **kwargs
                )
                for sys in spin_sys
            )
            amp = Parallel(
                n_jobs=n_jobs,
                verbose=verbose,
                backend="loky",
                # **{
                #     "backend": {
                #         "threads": "threading",
                #         "processes": "multithreading",
                #         None: None,
                #     }["threads"]
                # },
            )(jobs)

            # self.indexes.append(indexes)

            gyromagnetic_ratio = method.channels[0].gyromagnetic_ratio
            B0 = method.spectral_dimensions[0].events[0].magnetic_flux_density
            origin_offset = np.abs(B0 * gyromagnetic_ratio * 1e6)
            for seq in method.spectral_dimensions:
                seq.origin_offset = origin_offset

            if isinstance(amp[0], list):
                simulated_data = []
                for item in amp:
                    simulated_data += item
            if isinstance(amp[0], np.ndarray):
                simulated_data = [np.asarray(amp).sum(axis=0)]

            method.simulation = (
                self._as_csdm_object(simulated_data, method)
                if pack_as_csdm
                else np.asarray(simulated_data)
            )

    def save(self, filename: str, with_units: bool = True):
        """Serialize the simulator object to a JSON file.

        Args:
            bool with_units: If true, the attribute values are serialized as physical
                quantities expressed as a string with a value and a unit. If false,
                the attribute values are serialized as floats.
            str filename: The filename of the serialized file.

        Example
        -------

        >>> sim.save('filename') # doctest: +SKIP
        """
        sim = self.json(True, True) if with_units else self.reduced_dict()
        with open(filename, "w", encoding="utf8") as outfile:
            json.dump(
                sim, outfile, ensure_ascii=False, sort_keys=False, allow_nan=False
            )

    @classmethod
    def load(cls, filename: str, parse_units: bool = True):
        """Load the :class:`~mrsimulator.Simulator` object from a JSON file by parsing.

        Args:
            bool parse_units: If true, parse the attribute values from the serialized
                file for physical quantities, expressed as a string with a value and a
                unit.
            str filename: The filename of a JSON serialized mrsimulator file.

        Returns:
            A :class:`~mrsimulator.Simulator` object.

        Example
        -------

        >>> sim_1 = sim.load('filename') # doctest: +SKIP

        .. seealso::
            :ref:`load_spin_systems`
        """
        val = import_json(filename)
        return Simulator.parse(val, parse_units)

    @classmethod
    def parse(cls, py_dict: dict, parse_units: bool = True):
        """Parse a dictionary for Simulator object.

        Args:
            dict py_dict: Dictionary object.
            bool parse_units: It true, parse quantity from string.
        """
        return (
            Simulator.parse_dict_with_units(py_dict)
            if parse_units
            else Simulator(**py_dict)
        )

    def sites(self):
        """Unique sites within the Simulator object as a list of Site objects.

        Returns:
            A :ref:`sites_api` object.

        Example
        -------

        >>> sites = sim.sites() # doctest: +SKIP
        """
        unique_sites = []
        for sys in self.spin_systems:
            for site in sys.sites:
                if site not in unique_sites:
                    unique_sites.append(site)

        return Sites(unique_sites)

    def _as_csdm_object(self, data: np.ndarray, method: Method) -> cp.CSDM:
        """Converts the simulation data from the given method to a CSDM object. Read
        `csdmpy <https://csdmpy.readthedocs.io/en/stable/>`_ for details

        Return:
            A CSDM object.
        """
        dim = [sd.to_csdm_dimension() for sd in method.spectral_dimensions[::-1]]
        _ = [
            item.to("ppm", "nmr_frequency_ratio")
            for item in dim
            if item.origin_offset != 0
        ]

        dv = [
            {
                "type": "internal",
                "quantity_type": "scalar",
                "numeric_type": "float64",
                "components": [datum],
                **self._get_dv_metadata(index),
            }
            for index, datum in enumerate(data)
        ]

        return cp.CSDM(dimensions=dim, dependent_variables=dv)

    def _get_dv_metadata(self, index):
        """Update the name and description of the dependent variable attributes
        using fields from the spin system."""
        if self.config.decompose_spectrum != "spin_system":
            return {}

        if self.spin_systems == []:
            return {}

        obj = {}
        label = self.spin_systems[index].label
        if label not in ["", None]:
            obj["components_label"] = [label]

        name = self.spin_systems[index].name
        name = name if name not in ["", None] else f"spin system {index}"
        obj["name"] = name

        description = self.spin_systems[index].description
        if description not in ["", None]:
            obj["description"] = description

        obj["application"] = {
            "com.github.DeepanshS.mrsimulator": {
                "spin_systems": [self.spin_systems[index].json()]
            }
        }
        return obj


class Sites(AbstractList):
    """A list of unique :ref:`site_api` objects within a simulator object."""

    def __init__(self, data=[]):
        super().__init__(data)
        euler = ["alpha", "beta", "gamma"]
        self.site_labels = [
            "name",
            "label",
            "description",
            "isotope",
            "isotropic_chemical_shift",
            *[f"shielding_symmetric.{_}" for _ in ["zeta", "eta", *euler]],
            *[f"quadrupolar.{_}" for _ in ["Cq", "eta", *euler]],
        ]

    def __setitem__(self, index, item):
        """Set an item to the list at index"""
        if isinstance(item, Site):
            self._list[index] = item
        elif isinstance(item, dict):
            self._list[index] = Site(**item)
        else:
            raise ValueError("Only object of type Site is allowed.")

    def to_pd(self):
        """Return sites as a pandas dataframe."""
        row = {item: [] for item in self.site_labels}
        sites = [item.json() for item in self._list]
        for site in sites:
            site_ = flatten_dict(site)
            keys = site_.keys()
            for item in self.site_labels:
                val = None if item not in keys else site_[item]
                row[item].append(val)

        row_len = len(row["name"])
        nones = [None] * row_len
        for item in self.site_labels:
            if row[item] == nones:
                row.pop(item)

        return pd.DataFrame(row)


def get_chunks(items_list, n_jobs):
    """Return the chucks of into list into roughly n_jobs equal chunks

    Args:
        (list) items_list: The input list to divide into n_jobs chunks.
        (int) n_jobs: Number of chunks of input list.
    """
    if n_jobs < 0:
        n_jobs += __CPU_count__ + 1
    list_len = len(items_list)
    n_blocks, n_left = list_len // n_jobs, list_len % n_jobs

    chunks = [0] + [n_blocks] * n_jobs
    for i in range(n_left):
        chunks[i + 1] += 1

    for i in range(1, n_jobs + 1):
        chunks[i] += chunks[i - 1]
    slices = [slice(chunks[i], chunks[i + 1], None) for i in range(n_jobs)]
    return [items_list[item] for item in slices]



# -*- coding: utf-8 -*-
from copy import deepcopy
from typing import ClassVar
from typing import Dict
from typing import List

import numpy as np
from mrsimulator.utils.parseable import Parseable
from pydantic import Field

from .frequency_contrib import default_freq_contrib
from .frequency_contrib import freq_default
from .frequency_contrib import freq_list_all
from .frequency_contrib import FrequencyEnum
from .transition_query import TransitionQuery

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class Event(Parseable):
    r"""Base Event class defines the spin environment and the transition query for a
    segment of the transition pathway.

    Attributes
    ----------

    fraction:
        The weight of the frequency contribution from the event. The default is 1.

    magnetic_flux_density:
        The macroscopic magnetic flux density, :math:`H_0`, of the applied external
        magnetic field during the event in units of T. The default value is ``9.4``.

    rotor_frequency:
        The sample spinning frequency :math:`\nu_r`, during the event in units of Hz.
        The default value is ``0``.

    rotor_angle:
        The angle between the sample rotation axis and the applied external magnetic
        field vector, :math:`\theta`, during the event in units of rad.
        The default value is ``0.9553166``, i.e. the magic angle.

    freq_contrib:
        A list of FrequencyEnum enumeration. The default is all frequency enumerations.

    transition_query:
        A TransitionQuery or an equivalent dict object listing the queries used in
        selecting the active transitions during the event. Only the active transitions
        from this query will contribute to the net frequency.
    """

    fraction: float = 1.0
    magnetic_flux_density: float = Field(default=9.4, ge=0)
    rotor_frequency: float = Field(default=0.0, ge=0)
    # 54.735 degrees = 0.9553166 radians
    rotor_angle: float = Field(default=0.955316618, ge=0, le=1.5707963268)
    freq_contrib: List[FrequencyEnum] = default_freq_contrib
    transition_query: TransitionQuery = TransitionQuery()
    # user_variables: List = None

    property_unit_types: ClassVar = {
        "magnetic_flux_density": "magnetic flux density",
        "rotor_frequency": "frequency",
        "rotor_angle": "angle",
    }

    property_default_units: ClassVar = {
        "magnetic_flux_density": "T",
        "rotor_frequency": "Hz",
        "rotor_angle": "rad",
    }

    property_units: Dict = {
        "magnetic_flux_density": "T",
        "rotor_frequency": "Hz",
        "rotor_angle": "rad",
    }

    class Config:
        validate_assignment = True

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """Parse the physical quantities of an Event object from a python dictionary
        object.

        Args:
            dict py_dict: Dict object
        """
        py_dict_copy = deepcopy(py_dict)
        return super().parse_dict_with_units(py_dict_copy)

    def json(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object, where
        the attribute value with physical quantity is expressed as a string with a
        value and a unit."""
        dict_ = super().json()
        # if "user_variables" in dict_.keys():
        #     dict_.pop("user_variables")
        if dict_["fraction"] == 1.0:
            dict_.pop("fraction")
        if dict_["freq_contrib"] == freq_default:
            dict_.pop("freq_contrib")
        return dict_

    def _freq_contrib_flags(self) -> np.ndarray:
        lst_ = [item.index() for item in self.freq_contrib]
        array = np.zeros(len(freq_list_all), dtype=int)
        array[lst_] = 1
        return array



# -*- coding: utf-8 -*-
from os import path

import numpy as np
import pytest
from monty.serialization import loadfn
from mrsimulator.method import Method
from mrsimulator.methods import BlochDecaySpectrum
from mrsimulator.methods import Method1D
from mrsimulator.methods import Method2D
from mrsimulator.methods import ThreeQ_VAS

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

MODULE_DIR = path.dirname(path.abspath(__file__))
TESTDATA = loadfn(path.join(MODULE_DIR, "test_data.json"))

MAS_DIM = {
    "count": 128,
    "spectral_width": 5e4,  # in Hz
    "reference_offset": 0,  # in Hz
    "events": [
        {
            "rotor_angle": 54.735 * np.pi / 180,
            "transition_query": {"P": [-1], "D": [0]},
        },
    ],
}


def test_more_spectral_dimensions():
    error = "The method allows 1 spectral dimension"
    with pytest.raises(ValueError, match=f".*{error}.*"):
        BlochDecaySpectrum(spectral_dimensions=[{}, {}])


def test_01():
    error = "method requires exactly 1 channel"
    with pytest.raises(ValueError, match=f".*{error}.*"):
        BlochDecaySpectrum(channels=["1H", "29Si"])


def test_02():
    e = "`rotor_frequency` value cannot be modified for Method2D method."
    with pytest.raises(ValueError, match=f".*{e}.*"):
        Method2D(channels=["87Rb"], rotor_frequency=10, spectral_dimensions=[{}, {}])


def test_03():
    """generic method declaration"""
    mth = Method2D(
        channels=["87Rb"],
        magnetic_flux_density=9.4,  # in T
        spectral_dimensions=[
            {"count": 1024, "spectral_width": 5e4},
            {"count": 1024, "spectral_width": 5e4},
        ],
    )

    assert TESTDATA["generic"] == mth.json()
    assert Method.parse_dict_with_units(mth.json()) == mth


def test_method_1D():
    error = "Expecting a 1x1 affine matrix."
    with pytest.raises(ValueError, match=f".*{error}.*"):
        Method1D(spectral_dimensions=[{}], affine_matrix=[1, 2, 3, 4])

    error = r"The method allows 1 spectral dimension\(s\), 2 given."
    with pytest.raises(ValueError, match=f".*{error}.*"):
        Method1D(spectral_dimensions=[{}, {}])

    # parse dict with units test
    dict_1d = {
        "channels": ["87Rb"],
        "magnetic_flux_density": "7 T",  # in T
        "rotor_angle": "54.735 deg",
        "rotor_frequency": "1e9 Hz",
        "spectral_dimensions": [
            {
                "count": 1024,
                "spectral_width": "10 kHz",  # in Hz
                "reference_offset": "-4 kHz",  # in Hz
                "events": [
                    {"fraction": 27 / 17, "freq_contrib": ["Quad2_0"]},
                    {"fraction": 1, "freq_contrib": ["Quad2_4"]},
                ],
            }
        ],
    }
    method1a = Method1D.parse_dict_with_units(dict_1d)
    assert Method.parse_dict_with_units(method1a.json()) == method1a

    method1b = Method1D(
        channels=["87Rb"],
        magnetic_flux_density=7,  # in T
        rotor_angle=54.735 * np.pi / 180,
        rotor_frequency=1e9,
        spectral_dimensions=[
            {
                "count": 1024,
                "spectral_width": 1e4,  # in Hz
                "reference_offset": -4e3,  # in Hz
                "events": [
                    {"fraction": 27 / 17, "freq_contrib": ["Quad2_0"]},
                    {"fraction": 1, "freq_contrib": ["Quad2_4"]},
                ],
            }
        ],
    )

    assert method1a == method1b


def test_method_2D():

    error = "The first element of the affine matrix cannot be zero."
    with pytest.raises(ValueError, match=f".*{error}.*"):
        Method2D(spectral_dimensions=[{}, {}], affine_matrix=[0, 1, 2, 3])

    # parse dict with units test
    dict_1d = {
        "channels": ["87Rb"],
        "magnetic_flux_density": "7 T",  # in T
        "rotor_angle": "54.735 deg",
        "spectral_dimensions": [
            {
                "count": 1024,
                "spectral_width": "10 kHz",  # in Hz
                "events": [
                    {"fraction": 27 / 17, "freq_contrib": ["Quad2_0"]},
                    {"fraction": 1, "freq_contrib": ["Quad2_4"]},
                ],
            },
            {
                "count": 1024,
                "spectral_width": "10 kHz",  # in Hz
                "events": [
                    {"fraction": 27 / 17, "freq_contrib": ["Quad2_0"]},
                    {"fraction": 1, "freq_contrib": ["Quad2_4"]},
                ],
            },
        ],
    }
    method1a = Method2D.parse_dict_with_units(dict_1d)
    assert Method.parse_dict_with_units(method1a.json()) == method1a

    method1b = Method2D(
        channels=["87Rb"],
        magnetic_flux_density=7,  # in T
        rotor_angle=54.735 * np.pi / 180,
        spectral_dimensions=[
            {
                "count": 1024,
                "spectral_width": 1e4,  # in Hz
                "events": [
                    {"fraction": 27 / 17, "freq_contrib": ["Quad2_0"]},
                    {"fraction": 1, "freq_contrib": ["Quad2_4"]},
                ],
            },
            {
                "count": 1024,
                "spectral_width": 1e4,  # in Hz
                "events": [
                    {"fraction": 27 / 17, "freq_contrib": ["Quad2_0"]},
                    {"fraction": 1, "freq_contrib": ["Quad2_4"]},
                ],
            },
        ],
    )

    assert method1a == method1b


def test_04():
    """SAS method declaration"""
    mth = Method2D(
        channels=["87Rb"],
        magnetic_flux_density=9.4,  # in T
        spectral_dimensions=[
            {
                "count": 512,
                "spectral_width": 5e4,  # in Hz
                "events": [
                    {
                        "rotor_angle": 70.12 * np.pi / 180,
                        "transition_query": {"P": [-1], "D": [0]},
                    },
                ],
            },
            MAS_DIM.copy(),
        ],
    )

    assert TESTDATA["SAS"] == mth.json()
    assert Method.parse_dict_with_units(mth.json()) == mth
    assert Method2D.parse_dict_with_units(mth.json()) == mth


def test_BlochDecaySpectrum():
    # test-1
    m1 = BlochDecaySpectrum()

    dimension_dictionary_ = {
        "count": 1024,
        "spectral_width": "25000.0 Hz",
        "reference_offset": "0.0 Hz",
    }

    should_be = {
        "name": "BlochDecaySpectrum",
        "channels": ["1H"],
        "magnetic_flux_density": "9.4 T",
        "rotor_frequency": "0.0 Hz",
        "rotor_angle": "0.955316618 rad",
        "spectral_dimensions": [dimension_dictionary_],
    }
    dict_ = m1.json()
    assert Method.parse_dict_with_units(dict_) == m1

    dict_.pop("description")
    assert dict_ == should_be

    # test-2
    m2_dict = {
        "channels": ["29Si"],
        "magnetic_flux_density": "11.7 T",
        "rotor_angle": "90 deg",
        "spectral_dimensions": [{}],
    }
    m2 = BlochDecaySpectrum.parse_dict_with_units(m2_dict)

    angle = 90 * np.pi / 180
    dimension_dictionary_ = {
        "count": 1024,
        "spectral_width": "25000.0 Hz",
        "reference_offset": "0.0 Hz",
    }

    should_be = {
        "name": "BlochDecaySpectrum",
        "channels": ["29Si"],
        "magnetic_flux_density": "11.7 T",
        "rotor_frequency": "0.0 Hz",
        "rotor_angle": f"{angle} rad",
        "spectral_dimensions": [dimension_dictionary_],
    }

    dict_ = m2.json()
    assert Method.parse_dict_with_units(dict_) == m2

    dict_.pop("description")
    assert dict_ == should_be


def test_05():
    """Satellite to central correlation method declaration"""
    sp0 = dict(
        count=512,
        spectral_width=5e6,
        events=[
            {
                "rotor_angle": 0 * np.pi / 180,
                "transition_query": {"P": [-1], "D": [2, -2]},
            }
        ],
    )
    mth = Method2D(
        channels=["87Rb"],
        magnetic_flux_density=9.4,  # in T
        spectral_dimensions=[sp0, MAS_DIM.copy()],
    )

    assert TESTDATA["STMAS"] == mth.json()
    assert Method.parse_dict_with_units(TESTDATA["STMAS"]) == mth
    assert Method2D.parse_dict_with_units(TESTDATA["STMAS"]) == mth


def test_3QMAS():
    """3Q MAS correlation method declaration"""
    mth = ThreeQ_VAS(
        channels=["87Rb"],
        magnetic_flux_density=9.4,  # in T
        spectral_dimensions=[
            {"count": 512, "spectral_width": 5e6},
            {"count": 128, "spectral_width": 5e4},
        ],
    )

    assert np.allclose(mth.affine_matrix, [0.5625, 0.4375, 0, 1])
    assert ThreeQ_VAS.parse_dict_with_units(mth.json()) == mth


def test_06():
    """Test with order"""
    test_03()
    test_05()
    test_04()


def test_methods():
    das = Method2D(
        channels=["87Rb"],
        magnetic_flux_density=4.2,  # in T
        spectral_dimensions=[
            {
                "count": 256,
                "spectral_width": 2e4,  # in Hz
                "reference_offset": -5e3,  # in Hz
                "label": "70.12 dimension",
                "events": [
                    {
                        "rotor_angle": 70.12 * 3.14159 / 180,  # in rads
                        "transition_query": {"P": [-1], "D": [0]},
                    },
                    {
                        "rotor_angle": 30.12 * 3.14159 / 180,  # in rads
                        "transition_query": {"P": [-1], "D": [0]},
                    },
                ],
            },
            # The last spectral dimension block is the direct-dimension
            {
                "count": 256,
                "spectral_width": 3e4,  # in Hz
                "reference_offset": -7e3,  # in Hz
                "label": "MAS dimension",
                "events": [
                    {
                        "rotor_angle": 54.735 * 3.14159 / 180,  # in rads
                        "transition_query": {"P": [-1], "D": [0]},
                    }
                ],
            },
        ],
    )

    assert das.affine_matrix is None
    assert das.json() == TESTDATA["DAS"]
    assert Method.parse_dict_with_units(das.json()) == das
    assert Method2D.parse_dict_with_units(das.json()) == das



# -*- coding: utf-8 -*-
import numpy as np

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


def get_principal_components(zeta, eta):
    """Return the principal components of a traceless second-rank symmetric
    Cartesian tensor.

    Args:
        zeta: The zeta parameter in PAS, according to the Haeberlen convention.
        eta: The eta parameter in PAS, according to the Haeberlen convention.
    """
    xx = -0.5 * zeta * (eta + 1.0)
    yy = 0.5 * zeta * (eta - 1.0)
    zz = zeta

    return [xx, yy, zz]


def get_Haeberlen_components(tensors):
    """Return zeta and eta parameters of the tensor using the Haeberlen convention.

    Args:
        ndarray tensors: A `N x 3 x 3` ndarray of `N` traceless symmetric second-rank
            Cartesian tensors.
    """
    n = tensors.shape[0]
    eig_val = np.linalg.eigvalsh(tensors)
    eig_val_sort_ = np.argsort(np.abs(eig_val), axis=1, kind="mergesort")
    eig_val_sort_ = (eig_val_sort_.T + 3 * np.arange(n)).T.ravel()
    eig_val_sorted = eig_val.ravel()[eig_val_sort_].reshape(n, 3)

    eig_val_sort_ = eig_val = None
    del eig_val_sort_, eig_val

    zeta = eig_val_sorted[:, -1]
    eta = (eig_val_sorted[:, 0] - eig_val_sorted[:, 1]) / zeta

    return zeta, eta


def x_y_from_zeta_eta(zeta, eta):
    """Convert the zeta, eta coordinates from the Haeberlen convention to the
    x-y notation."""
    xa = np.empty(zeta.size)
    ya = np.empty(zeta.size)

    index = np.where(zeta >= 0)
    temp = np.tan(0.7853981634 * eta[index])
    ya[index] = np.sqrt(zeta[index] * zeta[index] / (temp * temp + 1.0))
    xa[index] = temp * ya[index]

    index = np.where(zeta < 0)
    temp = np.tan(0.7853981634 * eta[index])
    xa[index] = np.sqrt(zeta[index] * zeta[index] / (temp * temp + 1.0))
    ya[index] = temp * xa[index]

    zeta = eta = None
    del zeta, eta

    return xa, ya


def x_y_to_zeta_eta(x, y):
    """Same as def x_y_to_zeta_eta, but for ndarrays."""
    x = np.abs(x)
    y = np.abs(y)
    zeta = np.sqrt(x ** 2 + y ** 2)  # + offset
    eta = np.ones(zeta.shape)
    index = np.where(x > y)
    zeta[index] = -zeta[index]
    eta[index] = (4.0 / np.pi) * np.arctan(y[index] / x[index])

    index = np.where(x < y)
    eta[index] = (4.0 / np.pi) * np.arctan(x[index] / y[index])

    return zeta.ravel(), eta.ravel()



# -*- coding: utf-8 -*-
from copy import deepcopy
from os import path
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Union

import csdmpy as cp
import numpy as np
from monty.serialization import loadfn
from mrsimulator.spin_system.isotope import Isotope
from mrsimulator.transition import Transition
from mrsimulator.transition import TransitionPathway
from mrsimulator.utils.parseable import Parseable
from pydantic import validator

from .event import Event
from .spectral_dimension import SpectralDimension
from .utils import cartesian_product
from .utils import D_symmetry_indexes
from .utils import expand_spectral_dimension_object
from .utils import P_symmetry_indexes
from .utils import query_permutations

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

MODULE_DIR = path.dirname(path.abspath(__file__))
NAMED_METHODS = loadfn(path.join(MODULE_DIR, "named_methods.json"))["named_methods"]


class Method(Parseable):
    r"""Base Method class. A method class represents the NMR method.

    Attributes
    ----------

    channels: List[str] (optional).
        The value is a list of isotope symbols over which the given method applies.
        An isotope symbol is given as a string with the atomic number followed by its
        atomic symbol, for example, '1H', '13C', and '33S'. The default is an empty
        list.
        The number of isotopes in a `channel` depends on the method. For example, a
        `BlochDecaySpectrum` method is a single channel method, in which case, the
        value of this attribute is a list with a single isotope symbol, ['13C'].

        Example
        -------

        >>> bloch = Method()
        >>> bloch.channels = ['1H']

    spectral_dimensions: List[:ref:`spectral_dim_api`] or List[dict] (optional).
        The number of spectral dimensions depends on the given method. For example, a
        `BlochDecaySpectrum` method is a one-dimensional method and thus requires a
        single spectral dimension. The default is a single default
        :ref:`spectral_dim_api` object.

        Example
        -------

        >>> bloch = Method()
        >>> bloch.spectral_dimensions = [SpectralDimension(count=8, spectral_width=50)]
        >>> # or equivalently
        >>> bloch.spectral_dimensions = [{'count': 8, 'spectral_width': 50}]

    simulation: CSDM or ndarray (N/A).
        An object holding the result of the simulation. The initial value of this
        attribute is None. A value is assigned to this attribute when you run the
        simulation using the :meth:`~mrsimulator.Simulator.run` method.

    experiment: CSDM or ndarray (optional).
        An object holding the experimental measurement for the given method, if
        available. The default value is None.

        Example
        -------

        >>> bloch.experiment = my_data # doctest: +SKIP

    name: str (optional).
        Name or id of the method. The default value is None.

        Example
        -------

        >>> bloch.name = 'BlochDecaySpectrum'
        >>> bloch.name
        'BlochDecaySpectrum'

    label: str (optional).
        Label for the method. The default value is None.

        Example
        -------

        >>> bloch.label = 'One pulse acquired spectrum'
        >>> bloch.label
        'One pulse acquired spectrum'

    description: str (optional).
        A description of the method. The default value is None.

        Example
        -------

        >>> bloch.description = 'Huh!'
        >>> bloch.description
        'Huh!'

    affine_matrix: np.ndarray or 2D list (optional)
        A (`n` x `n`) affine transformation matrix, where `n` is the number of
        spectral_dimensions. If provided, the corresponding affine transformation is
        applied to the computed frequencies. The default is None, i.e., no
        transformation is applied.

        Example
        -------

        >>> method = Method2D()
        >>> method.affine_matrix = [[1, -1], [0, 1]]
        >>> print(method.affine_matrix)
        [[ 1 -1]
         [ 0  1]]
    """
    name: str = None
    label: str = None
    description: str = None
    channels: List[str] = []
    spectral_dimensions: List[SpectralDimension] = [SpectralDimension()]
    affine_matrix: Union[np.ndarray, List] = None
    simulation: Union[cp.CSDM, np.ndarray] = None
    experiment: Union[cp.CSDM, np.ndarray] = None

    property_default_units: ClassVar = {
        "magnetic_flux_density": "T",
        "rotor_angle": "rad",
        "rotor_frequency": "Hz",
    }

    property_units: Dict = {
        "magnetic_flux_density": "T",
        "rotor_angle": "rad",
        "rotor_frequency": "Hz",
    }

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    def __eq__(self, other):
        if not isinstance(other, Method):
            return False
        check = [
            self.name == other.name,
            self.label == other.label,
            self.description == other.description,
            self.channels == other.channels,
            self.spectral_dimensions == other.spectral_dimensions,
            np.all(self.affine_matrix == other.affine_matrix),
            self.simulation == other.simulation,
            self.experiment == other.experiment,
        ]
        if np.all(check):
            return True
        return False

    @validator("channels", always=True)
    def validate_channels(cls, v, *, values, **kwargs):
        return [Isotope(symbol=_) for _ in v]

    @staticmethod
    def __check_csdm__(data):
        if data is None:
            return None
        if isinstance(data, dict):
            return cp.parse_dict(data)
        if isinstance(data, cp.CSDM):
            return data
        raise ValueError("Unable to read the data.")

    @validator("experiment", pre=True, always=True)
    def validate_experiment(cls, v, *, values, **kwargs):
        return cls.__check_csdm__(v)

    @validator("simulation", pre=True, always=True)
    def validate_simulation(cls, v, *, values, **kwargs):
        if isinstance(v, np.ndarray):
            return v
        return cls.__check_csdm__(v)

    @validator("affine_matrix", pre=True, always=True)
    def validate_affine_matrix(cls, v, *, values, **kwargs):
        if v is None:
            return None
        v = np.asarray(v)
        dim_len = len(values["spectral_dimensions"])
        if v.size != dim_len ** 2:
            raise ValueError(f"Expecting a {dim_len}x{dim_len} affine matrix.")
        if v.ravel()[0] == 0:
            raise ValueError("The first element of the affine matrix cannot be zero.")
        return v

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """Parse the physical quantity from a dictionary representation of the Method
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict py_dict: A python dict representation of the Method object.

        Returns:
            A :ref:`method_api` object.
        """
        py_dict_copy = deepcopy(py_dict)

        if "spectral_dimensions" in py_dict_copy:
            py_dict_copy = expand_spectral_dimension_object(py_dict_copy)
            py_dict_copy["spectral_dimensions"] = [
                SpectralDimension.parse_dict_with_units(s)
                for s in py_dict_copy["spectral_dimensions"]
            ]

        if "simulation" in py_dict_copy:
            if py_dict_copy["simulation"] is not None:
                py_dict_copy["simulation"] = cp.parse_dict(py_dict_copy["simulation"])
        if "experiment" in py_dict_copy:
            if py_dict_copy["experiment"] is not None:
                py_dict_copy["experiment"] = cp.parse_dict(py_dict_copy["experiment"])

        return super().parse_dict_with_units(py_dict_copy)

    def update_spectral_dimension_attributes_from_experiment(self):
        """Update the spectral dimension attributes of the method to match the
        attributes of the experiment from the :attr:`~mrsimulator.Method.experiment`
        attribute."""
        spectral_dims = self.spectral_dimensions
        for i, dim in enumerate(self.experiment.dimensions):
            spectral_dims[i].count = dim.count
            spectral_dims[i].spectral_width = dim.count * dim.increment.to("Hz").value
            spectral_dims[i].reference_offset = dim.coordinates_offset.to("Hz").value
            spectral_dims[i].origin_offset = dim.origin_offset.to("Hz").value

    def dict(self, **kwargs):
        mth = super().dict(**kwargs)
        if self.simulation is not None:
            mth["simulation"] = self.simulation.to_dict(update_timestamp=True)
        if self.experiment is not None and isinstance(self.experiment, cp.CSDM):
            mth["experiment"] = self.experiment.to_dict()
        return mth

    def json(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object, where
        the attribute value with physical quantity is expressed as a string with a
        value and a unit.

        Returns:
            A python dict object.
        """

        mth = {_: self.__getattribute__(_) for _ in ["name", "label", "description"]}
        mth["channels"] = [item.json() for item in self.channels]
        mth["spectral_dimensions"] = [item.json() for item in self.spectral_dimensions]

        # add global parameters
        ev0 = self.spectral_dimensions[0].events[0]
        evt_d = Event.property_default_units
        global_ = {k: f"{ev0.__getattribute__(k)} {u}" for k, u in evt_d.items()}
        mth.update(global_)

        named = True if mth["name"] in NAMED_METHODS else False
        for dim in mth["spectral_dimensions"]:
            for ev in dim["events"]:
                # remove event objects with global values.
                _ = [ev.pop(k) if ev[k] == v else 0 for k, v in global_.items()]

                # remove transition query objects for named methods
                _ = ev.pop("transition_query") if named else 0

            if dim["events"] == [{} for _ in dim["events"]]:
                dim.pop("events")

        afm = self.affine_matrix
        mth["affine_matrix"] = None if afm is None else afm.tolist()

        sim = self.simulation
        mth["simulation"] = None if sim is None else sim.to_dict(update_timestamp=True)

        exp = self.experiment
        mth["experiment"] = None if exp is None else exp.to_dict()

        _ = [mth.pop(item) for item in [k for k, v in mth.items() if v is None]]
        return mth

    # def _simplify_events_json(self, py_dict):
    #     named = True if py_dict["name"] in NAMED_METHODS else False
    #     for dim in py_dict["spectral_dimensions"]:
    #         for ev in dim["events"]:
    #             # remove event objects with global values.
    #             for key, val in zip(list_g, global_):
    #                 _ = ev.pop(key) if ev[key] == val else 0

    #             # remove transition query objects for named methods
    #             _ = ev.pop("transition_query") if named else 0

    #         if dim["events"] == [{} for _ in range(len(dim["events"]))]:
    #             dim.pop("events")

    # def _get_symmetry_pathways(self, spin_system):
    #     list_of_P = []
    #     list_of_D = []
    #     for dim in self.spectral_dimensions:
    #         for ent in dim.events:
    #             list_of_P.append(
    #                 query_permutations(
    #                     ent.transition_query.dict(),
    #                     isotope=spin_system.get_isotopes(symbol=True),
    #                     channel=[item.symbol for item in self.channels],
    #                 )
    #             )
    #             if ent.transition_query.D is not None:
    #                 list_of_D.append(
    #                     query_permutations(
    #                         ent.transition_query.dict(),
    #                         isotope=spin_system.get_isotopes(symbol=True),
    #                         channel=[item.symbol for item in self.channels],
    #                         transition_symmetry="D",
    #                     )
    #                 )

    #     return {"P": list_of_P, "D": list_of_D}

    def _get_transition_pathways(self, spin_system):
        all_transitions = spin_system._all_transitions()

        segments = []
        for dim in self.spectral_dimensions:
            for ent in dim.events:
                # query the transitions for P symmetry
                selected_transitions = all_transitions[:]
                list_of_P = query_permutations(
                    ent.transition_query.dict(),
                    isotope=spin_system.get_isotopes(symbol=True),
                    channel=[item.symbol for item in self.channels],
                )
                indexes = P_symmetry_indexes(selected_transitions, list_of_P)
                selected_transitions = selected_transitions[indexes]

                # query the transitions for D symmetry
                if ent.transition_query.D is not None:
                    list_of_D = query_permutations(
                        ent.transition_query.dict(),
                        isotope=spin_system.get_isotopes(symbol=True),
                        channel=[item.symbol for item in self.channels],
                        transition_symmetry="D",
                    )
                    indexes = D_symmetry_indexes(selected_transitions, list_of_D)
                    selected_transitions = selected_transitions[indexes]

                segments += [selected_transitions]
        return segments

    def _get_transition_pathways_np(self, spin_system):
        segments = self._get_transition_pathways(spin_system)
        segments_index = [np.arange(item.shape[0]) for item in segments]
        cartesian_index = cartesian_product(*segments_index)
        return [
            [segments[i][j] for i, j in enumerate(item)] for item in cartesian_index
        ]

    def get_transition_pathways(self, spin_system) -> List[TransitionPathway]:
        """Return a list of transition pathways from the given spin system that satisfy
        the query selection criterion of the method.

        Args:
            SpinSystem spin_system: A SpinSystem object.

        Returns:
            An array of :ref:`transition_pathway_api` objects. Each TransitionPathway
            object is an ordered collection of Transition objects.

        Example:
            >>> from mrsimulator import SpinSystem
            >>> from mrsimulator.methods import ThreeQ_VAS
            >>> sys = SpinSystem(sites=[{'isotope': '27Al'}, {'isotope': '29Si'}])
            >>> method = ThreeQ_VAS(channels=['27Al'])
            >>> pprint(method.get_transition_pathways(sys))
            [|1.5, -0.5⟩⟨-1.5, -0.5| ⟶ |-0.5, -0.5⟩⟨0.5, -0.5|,
             |1.5, -0.5⟩⟨-1.5, -0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|,
             |1.5, 0.5⟩⟨-1.5, 0.5| ⟶ |-0.5, -0.5⟩⟨0.5, -0.5|,
             |1.5, 0.5⟩⟨-1.5, 0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|]
        """
        segments = self._get_transition_pathways_np(spin_system)
        return [
            TransitionPathway(
                [
                    Transition(initial=tr[0].tolist(), final=tr[1].tolist())
                    for tr in item
                ]
            )
            for item in segments
        ]

    def shape(self) -> tuple:
        """The shape of the method's spectral dimension array.

        Returns:
            tuple

        Example:
            >>> from mrsimulator.methods import Method2D
            >>> method = Method2D(spectral_dimensions=[{'count': 40}, {'count': 10}])
            >>> method.shape()
            (40, 10)
        """
        return tuple([item.count for item in self.spectral_dimensions])



# -*- coding: utf-8 -*-
"""The Operation class."""
from sys import modules
from typing import ClassVar

import numpy as np
from mrsimulator.utils.parseable import Parseable

__author__ = "Maxwell C. Venetos"
__email__ = "maxvenetos@gmail.com"


class Operation(Parseable):
    """A base class for signal processing operations."""

    module_name: ClassVar = None

    @property
    def function(self):
        return self.__class__.__name__

    def json(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object, where
        the attribute value with physical quantity is expressed as a string with a
        value and a unit."""
        my_dict = super().json()
        my_dict["function"] = self.function
        if hasattr(self, "type"):
            my_dict["type"] = self.type
        return my_dict

    @staticmethod
    def _get_dv_indexes(indexes, n):
        """Return a list of dependent variable indexes.

        Args:
            indexes: An interger, list of integers, or None indicating the dv indexes.
            n: Total number of dependent variables in the CSDM object.
        """
        if indexes is None:
            return np.arange(n)
        if isinstance(indexes, int):
            return [indexes]
        if isinstance(indexes, (list, tuple)):
            return np.asarray(indexes)

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """Parse dictionary for SignalProcessor

        Args:
            py_dict (dict): A python dictionary representation of the operation.
        """
        my_dict_copy = py_dict.copy()
        if "function" in my_dict_copy.keys():
            my_dict_copy.pop("function")
        return (
            super().parse_dict_with_units(my_dict_copy)
            if "type" not in my_dict_copy
            else getattr(modules[cls.module_name], my_dict_copy["type"])(**my_dict_copy)
        )

    @staticmethod
    def get_coordinates_in_units(x, unit):
        return x.to(unit).value

    @staticmethod
    def get_coordinates(dim):
        """Return the coordinates of dimension, dim, without equivalence units"""

        def get_coords(dim):
            equivalent_fn, equivalent_unit = dim._equivalencies, dim._equivalent_unit
            dim._equivalencies, dim._equivalent_unit = None, None

            coordinates = dim.coordinates

            dim._equivalencies, dim._equivalent_unit = equivalent_fn, equivalent_unit
            return coordinates

        return (
            get_coords(dim) if not hasattr(dim, "subtype") else get_coords(dim.subtype)
        )


class ModuleOperation(Operation):
    @property
    def type(self):
        """The type baseline function."""
        return self.__class__.__name__



# -*- coding: utf-8 -*-
"""Utility function module."""
import json
from urllib.parse import urlparse

from csdmpy.dependent_variable.download import download_file_from_url


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


def import_json(filename):
    res = urlparse(filename)
    if res[0] not in ["file", ""]:
        filename = download_file_from_url(filename)
    with open(filename, "rb") as f:
        content = f.read()
        return json.loads(str(content, encoding="UTF-8"))



# -*- coding: utf-8 -*-
#
#  test.pxd
#
#  @copyright Deepansh J. Srivastava, 2019-2021.
#  Created by Deepansh J. Srivastava.
#  Contact email = srivastava.89@osu.edu
#

from libcpp cimport bool as bool_t

cdef extern from "tables/trig.h":
    void generate_table()

cdef extern from "angular_momentum.h":
    void wigner_d_matrices(const int l, const int n, const double *angle, double *wigner)

    void wigner_d_matrices_from_exp_I_beta(const int l, const int n, const bool_t half,
                            const double complex *exp_I_beta, double *wigner)

    # void __wigner_rotation(const int l, const int n, const double *wigner, const double *cos_alpha,
    #                        const double complex *R_in, double complex *R_out)

    void __wigner_rotation_2(const int l, const int n, const double *wigner,
                            const void *exp_Im_alpha, const void *R_in, void *R_out)

    void single_wigner_rotation(const int l, const double *euler_angles, const void *R_in,
                            void *R_out)

    void wigner_dm0_vector(const int l, const double beta, double *R_out)

    void get_exp_Im_alpha(const unsigned int octant_orientations,
                            const bool_t allow_fourth_rank, void *exp_Im_alpha)

    void __batch_wigner_rotation(const unsigned int octant_orientations,
                            const unsigned int n_octants, double *wigner_2j_matrices, void *R2,
                            double *wigner_4j_matrices, void *R4, void *exp_Im_alpha, void *w2,
                            void *w4)


cdef extern from "octahedron.h":
    void averaging_setup(
        int nt,
        double complex *exp_I_alpha,
        double complex *exp_I_beta,
        double *amp)

cdef extern from "interpolation.h":
    void triangle_interpolation1D(
        double *freq1,
        double *freq2,
        double *freq3,
        double *amp,
        double *spec,
        int *points)

    void triangle_interpolation2D(
        double *freq11,
        double *freq12,
        double *freq13,
        double *freq21,
        double *freq22,
        double *freq23,
        double *amp,
        double *spec,
        int m0,
        int m1)

    void octahedronInterpolation(
        double *spec,
        double *freq,
        int nt,
        double *amp,
        int stride,
        int m)

cdef extern from "mrsimulator.h":
    void get_sideband_phase_components(
        unsigned int number_of_sidebands,
        double spin_frequency,
        double *pre_phase)

#     ctypedef struct MRS_plan

#     MRS_plan *MRS_create_plan(
#         unsigned int integration_density,
#         int number_of_sidebands,
#         double rotor_frequency_in_Hz,
#         double rotor_angle_in_rad, double increment,
#         bool_t allow_fourth_rank)


cdef extern from "object_struct.h":
    ctypedef struct site_struct:
        int number_of_sites;                     # Number of sites
        float *spin;                             # The spin quantum number
        double *gyromagnetic_ratio;              # Larmor frequency (MHz)
        double *isotropic_chemical_shift_in_ppm; # Isotropic chemical shift (Hz)
        double *shielding_symmetric_zeta_in_ppm; # Nuclear shielding anisotropy (Hz)
        double *shielding_symmetric_eta;         # Nuclear shielding asymmetry parameter
        double *shielding_orientation;           # Nuclear shielding PAS to CRS euler angles (rad.)
        double *quadrupolar_Cq_in_Hz;            # Quadrupolar coupling constant (Hz)
        double *quadrupolar_eta;                 # Quadrupolar asymmetry parameter
        double *quadrupolar_orientation;         # Quadrupolar PAS to CRS euler angles (rad.)

cdef extern from "method.h":
    ctypedef struct MRS_event:
        double fraction                    # The weighted frequency contribution from the event.
        double magnetic_flux_density_in_T  #  he magnetic flux density in T.
        double rotor_angle_in_rad          # The rotor angle in radians.
        double rotor_frequency_in_Hz       # The sample rotation frequency in Hz.

    ctypedef struct MRS_dimension:
        int count                       #  The number of coordinates along the dimension.
        double increment                # Increment of coordinates along the dimension.
        double coordinates_offset       #  Start coordinate of the dimension.
        MRS_event *events               # Holds a list of events.
        unsigned int n_events           # The number of events.

    # MRS_dimension *MRS_create_dimensions(
    #     MRS_averaging_scheme *scheme,
    #     int count,
    #     double coordinates_offset,
    #     double increment,
    #     double *magnetic_flux_density_in_T,
    #     double *rotor_frequency_in_Hz,
    #     double *rotor_angle_in_rad,
    #     unsigned int n_events,
    #     int number_of_sidebands)

cdef extern from "simulation.h":
    void mrsimulator_core(
        # spectrum information and related amplitude
        double * spec,
        double spectral_start,
        double spectral_increment,
        int number_of_points,

        site_struct *sites,
        MRS_dimension *dimensions[],              # the dimensions in the method.

        int quad_second_order,                    # Quad theory for second order,
        bool_t remove_2nd_order_quad_isotropic,   # remove the isotropic contribution from the
                                                  # second order quad Hamiltonian.

        # spin rate, spin angle and number spinning sidebands
        unsigned int number_of_sidebands,
        double rotor_frequency_in_Hz,
        double rotor_angle_in_rad,

        # The transition as transition[0] = mi and transition[1] = mf
        double *transition,
        int integration_density,
        unsigned int integration_volume,      # 0-octant, 1-hemisphere, 2-sphere.
        bool_t interpolation
        )



# -*- coding: utf-8 -*-

IGNORE = ["simulation", "experiment"]


def _reduce_dict(dict_obj, exclude=["property_units"]):
    """Reduce the dict by removing all key-value pair corresponding to keys listed in
    the `exclude` argument and keys with value as None.

    Args:
        exclude: List of keys to exclude from the dict.
    Returns: A dict.
    """
    if isinstance(dict_obj, dict):
        obj = {}
        for k, v in dict_obj.items():
            if k in IGNORE and v is not None:
                obj[k] = v
                continue
            if k in exclude or v is None:
                continue
            if k == "isotope":
                obj[k] = v["symbol"]
                continue
            if k == "channels":
                obj[k] = [item["symbol"] for item in v]
                continue
            elif isinstance(v, dict):
                obj[k] = _reduce_dict(v)
            elif isinstance(v, list):
                obj[k] = [_reduce_dict(_) for _ in v]
            else:
                obj[k] = v
        return obj

    # if isinstance(dict_obj, list):
    #     obj = []
    #     for v in dict_obj:
    #         if v is None:
    #             continue
    #         if isinstance(v, dict):
    #             obj.append(_reduce_dict(v))
    #         elif isinstance(v, list):
    #             obj.append([_reduce_dict(_) for _ in v])
    #         else:
    #             obj.append(v)
    #     return obj

    return dict_obj



# -*- coding: utf-8 -*-
from functools import reduce
from itertools import permutations

import numpy as np

__author__ = ["Deepansh J. Srivastava", "Maxwell C. Venetos"]
__email__ = ["srivastava.89@osu.edu", "maxvenetos@gmail.com"]


def expand_spectral_dimension_object(py_dict):
    glb = {}
    list_g = ["magnetic_flux_density", "rotor_frequency", "rotor_angle"]
    for item in list_g:
        if item in py_dict.keys():
            glb[item] = py_dict[item]
    glb_keys = glb.keys()

    for dim in py_dict["spectral_dimensions"]:
        if "events" not in dim:
            dim["events"] = [{}]
        for ev in dim["events"]:
            intersect = set(ev.keys()).intersection(set(glb_keys))
            for k in glb:
                if k not in intersect:
                    ev[k] = glb[k]

    _ = [py_dict.pop(item) for item in glb_keys]

    return py_dict


def cartesian_product(*arrays):
    la = len(arrays)
    dtype = np.result_type(*arrays)
    arr = np.empty([len(a) for a in arrays] + [la], dtype=dtype)
    for i, a in enumerate(np.ix_(*arrays)):
        arr[..., i] = a
    return arr.reshape(-1, la)


def get_symmetry_indexes(fn, list_of_sym):
    n = fn.shape[1]
    return reduce(
        np.union1d,
        [
            reduce(np.intersect1d, [np.where(fn[:, i] == search[i]) for i in range(n)])
            for search in list_of_sym
        ],
        np.asarray([], dtype=np.int64),
    )


def P_symmetry_indexes(transitions, list_of_P):
    P_fn = transitions[:, 1, :] - transitions[:, 0, :]
    return get_symmetry_indexes(P_fn, list_of_P)


def D_symmetry_indexes(transitions, list_of_D):
    D_fn = transitions[:, 1, :] ** 2 - transitions[:, 0, :] ** 2
    return get_symmetry_indexes(D_fn, list_of_D)


def get_iso_dict(channel, isotope):
    """
    Parse the spin system sites to determine indices of each isotope that is part of
    the method channel.

    Args:
        channel: List object
        isotope: List object
    """
    intersection = set(isotope).intersection(set(channel))
    isotope = np.asarray(isotope)
    return {item: (np.where(isotope == item))[0] for item in intersection}


def query_permutations(query, isotope, channel, transition_symmetry="P"):
    """
    Determines the transition symmetries that are involved in a given transition query.

    Args:
        query: Transition query dict object
        channel: List of channels as atomic number followed by symbol. Eg. '29Si', '1H'
        isotope: List of isotopes within the spin system. Eg. ['29Si', '29Si', '1H']
        transition_symmetry: str object. Derived from a transition query
    """
    P_permutated = []

    iso_dict = get_iso_dict(channel=channel, isotope=isotope)
    query_symmetry = query[transition_symmetry]  # get the query for symmetry element.

    # def warn_message(id_):
    #     return (
    #         f"Channel asks for isotope `{id_}` but it is not present in the list of "
    #         "spin systems."
    #     )

    on_fail_message = (
        "The length of the transition query symmetry elements cannot exceed than the "
        "number of channels."
    )
    for i, channel_id in enumerate(channel):
        # Check if method's channel isotope is present in the spin system
        if channel_id not in iso_dict:
            # warnings.warn(warn_message(channel_id))
            return np.asarray([])

        n_sites_channel_i = iso_dict[channel_id].size
        channel_query = query_symmetry[f"channel-{i+1}"]

        temp_P = []
        for item in channel_query:
            query_item_len = len(item)
            # Check transition query exceed the number of isotopes present
            if query_item_len > n_sites_channel_i:
                raise ValueError(on_fail_message)

            item += (n_sites_channel_i - query_item_len) * [0]
            temp_P += list(set(permutations(item)))
        P_permutated += [temp_P]

    # Expand the permutation to the number of sites in the spin system
    permutation_length = max(len(item) for item in P_permutated)
    P_expanded = np.zeros((permutation_length, len(isotope)))
    for i, iso_trans_symmetry in enumerate(P_permutated):
        # Update the channel-i indexes with the permuted symmetry.
        P_expanded[:, iso_dict[channel[i]]] = iso_trans_symmetry

    return P_expanded



# -*- coding: utf-8 -*-
from copy import deepcopy
from os import path

from monty.serialization import loadfn
from mrsimulator.method import Method
from mrsimulator.method.event import Event
from mrsimulator.method.spectral_dimension import SpectralDimension


MODULE_DIR = path.dirname(path.abspath(__file__))
METHODS_DATA = loadfn(path.join(MODULE_DIR, "methods_data.json"))

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


default_units = {
    "magnetic_flux_density": "T",
    "rotor_frequency": "Hz",
    "rotor_angle": "rad",
}


# class namedMethod:
#     def __new__(cls, py_dict):
#         method_name = py_dict["name"]


def prepare_method_structure(template, **kwargs):
    keys = kwargs.keys()
    n_channels = template["number_of_channels"]
    name = template["name"] if "name" not in keys else kwargs["name"]
    desc = (
        template["description"] if "description" not in keys else kwargs["description"]
    )
    label = None if "label" not in keys else kwargs["label"]
    affine_matrix = None if "affine_matrix" not in keys else kwargs["affine_matrix"]

    simulation = None if "simulation" not in keys else kwargs["simulation"]
    experiment = None if "experiment" not in keys else kwargs["experiment"]

    prep = {
        "name": name,
        "description": desc,
        "label": label,
        "simulation": simulation,
        "experiment": experiment,
        "affine_matrix": affine_matrix,
    }

    if "channels" in kwargs:
        prep["channels"] = kwargs["channels"]
        given_n_channels = len(prep["channels"])
        if given_n_channels != n_channels:
            raise ValueError(
                f"The method requires exactly {n_channels} channel(s), "
                f"{given_n_channels} provided."
            )
        kwargs.pop("channels")
    else:
        prep["channels"] = ["1H" for _ in range(n_channels)]
    return prep


def parse_spectral_dimensions(spectral_dimensions, n):
    if spectral_dimensions == [{}]:
        return [{}] * n
    for dim in spectral_dimensions:
        if "events" in dim.keys():
            for evt in dim["events"]:
                parse_events(evt)
    return spectral_dimensions


def parse_events(evt):
    if "transition_query" not in evt.keys():
        return

    t_query = evt["transition_query"]
    if "P" in t_query.keys():
        t_query["P"] = parse_transition_query(t_query["P"])

    if "D" in t_query.keys():
        t_query["D"] = parse_transition_query(t_query["D"])


def parse_transition_query(query):
    if not isinstance(query, list):
        return query
    return {
        "channel-1": [[i] if not isinstance(i, (list, tuple)) else i for i in query]
    }


def generate_method_from_template(template, docstring=""):
    """Generate method object from json template."""
    # constructor
    def constructor(self, spectral_dimensions=[{}], **kwargs):
        # spectral_dimensions_root = deepcopy(spectral_dimensions)
        # kwargs_root = deepcopy(kwargs)
        if isinstance(spectral_dimensions[0], SpectralDimension):
            return Method(spectral_dimensions=spectral_dimensions, **kwargs)

        parse = False
        if "parse" in kwargs:
            parse = kwargs["parse"]
            kwargs.pop("parse")

        local_template = deepcopy(template)
        n_tem = len(local_template["spectral_dimensions"])

        spectral_dimensions = parse_spectral_dimensions(spectral_dimensions, n_tem)
        prep = prepare_method_structure(local_template, **kwargs)
        global_events = local_template["global_event_attributes"]
        ge = set(global_events)
        kw = set(kwargs)
        common = kw.intersection(ge)

        _check_rotor_frequency(common, prep["name"], kwargs)

        n_sp = len(spectral_dimensions)
        if n_tem < n_sp:
            raise ValueError(
                f"The method allows {n_tem} spectral dimension(s), {n_sp} given."
            )
        dim = []
        for i, s in enumerate(local_template["spectral_dimensions"]):
            events = _gen_event(s, spectral_dimensions[i], global_events, parse, kwargs)

            if "events" in spectral_dimensions[i]:
                spectral_dimensions[i].pop("events")

            params = {**spectral_dimensions[i], "events": events}
            params = params if parse else SpectralDimension(**params)
            dim.append(params)

        method = {**prep, "spectral_dimensions": dim}
        method = Method.parse_dict_with_units(method) if parse else Method(**method)
        return method

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """
        Parse the physical quantities of the method object from as a python dictionary.

        Args:
            py_dict: Dict object.
        """
        return cls.__new__(0, parse=True, **deepcopy(py_dict))

    description = template["description"]
    method = type(
        template["name"],
        (object,),
        {
            "__new__": constructor,
            "__str__": description,
            "__doc__": description + docstring,
            "parse_dict_with_units": parse_dict_with_units,
            "ndim": len(template["spectral_dimensions"]),
        },
    )
    return method


def _fix_strings_in_events(event):
    """Fix the default float values to string with units when parse dict with units
    function is called."""
    for key, val in event.items():
        if key in default_units and isinstance(val, float):
            unit = default_units[key]
            event[key] = f"{val} {unit}"
    return event


def _fill_missing_events_in_template(spectral_dimensions, s_template):
    """Fill the missing events in the template relative to the spectral dimensions."""
    if "events" not in spectral_dimensions:
        return

    s_tem_len = len(s_template["events"])
    sp_evt_len = len(spectral_dimensions["events"])
    if s_tem_len < sp_evt_len:
        diff = sp_evt_len - s_tem_len
        _ = [s_template["events"].append(Event().dict()) for _ in range(diff)]


def _check_rotor_frequency(common, name, kwargs):
    if common == set():
        return

    r_freq = kwargs["rotor_frequency"]
    if "rotor_frequency" in common and r_freq == "1000000000000.0 Hz":
        return

    info = "`, `".join(list(common))
    raise ValueError(f"`{info}` value cannot be modified for {name} method.")


def _gen_event(dim_template, dim_i, global_events, parse, kwargs):

    events = []
    _fill_missing_events_in_template(dim_i, dim_template)

    for j, e in enumerate(dim_template["events"]):
        kw = deepcopy(kwargs)
        ge = deepcopy(global_events)

        # Remove `property_units` from default events instance.
        if "property_units" in e:
            e.pop("property_units")

        if "events" in dim_i:
            kw.update(dim_i["events"][j])

        e.update(kw)
        # prioritize the keyword arguments over the global arguments.
        ge.update(kw)
        e.update(ge)

        params = _fix_strings_in_events(e) if parse else Event(**e)
        events.append(params)

    return events



# -*- coding: utf-8 -*-
from typing import ClassVar
from typing import Dict
from typing import Union

import numpy as np
from pydantic import validator

from ._base import ModuleOperation
from .utils import _str_to_quantity
from .utils import CONST

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"


class Baseline(ModuleOperation):
    module_name: ClassVar = __name__

    @property
    def function(self):
        return "baseline"


class Polynomial(Baseline):
    r"""Add a baseline polynomial to all dependent variables (y) in the CSDM object.

    The baseline function is

    .. math::
            f(x) = \sum_{i=0}^3 x_i \times x^i,

    where :math:`x` is the CSDM dimension at index ``dim_index``.

    Args:
        str x0: Constant zeroth order offset. The default value is 0.
        str x1: First order factor. The default value is 0.
        str x2: Second order factor. The default value is 0.
        str x3: Third order factor. The default value is 0.
        int dim_index: The index of the CSDM dimension along which the operation is
            applied. The default is the dimension at index 0.

    Example
    -------

    >>> from mrsimulator import signal_processing as sp
    >>> operation1 = sp.baseline.Polynomial(x0=20, x1=-10)
    """

    x0: Union[float, str] = 0
    x1: Union[float, str] = 0
    x2: Union[float, str] = 0
    x3: Union[float, str] = 0
    dim_index: int = 0
    property_units: Dict = {"x0": CONST, "x1": CONST, "x2": CONST, "x3": CONST}

    @validator("x0")
    def str_to_quantity0(cls, v, values):
        return _str_to_quantity(v, values, "x0")

    @validator("x1")
    def str_to_quantity1(cls, v, values):
        return _str_to_quantity(v, values, "x1")

    @validator("x2")
    def str_to_quantity2(cls, v, values):
        return _str_to_quantity(v, values, "x2")

    @validator("x3")
    def str_to_quantity3(cls, v, values):
        return _str_to_quantity(v, values, "x3")

    def operate(self, data):
        """Applies the operation.

        Args:
            data: CSDM object
        """
        x = data.dimensions[self.dim_index].coordinates
        if self.x1 == self.x2 == self.x3 == 0:
            data += self.x0
            return data

        if "x1" in self.property_units:
            unit = 1 * self.property_units["x1"]
        elif "x2" in self.property_units:
            unit = np.sqrt(1 * self.property_units["x2"]).unit
        elif "x3" in self.property_units:
            unit = np.cbrt(1 * self.property_units["x3"]).unit

        d1 = self.get_coordinates_in_units(x, unit=1.0 / unit)
        d2 = d1 ** 2
        fn = self.x0 + self.x1 * d1 + d2 * (self.x2 + d1 * self.x3)
        for item in data.y:
            item.components += fn
        return data


class ConstantOffset(Baseline):
    r"""Add an offset to the dependent variables (y) of the CSDM object.

    .. math::
        y += \text{offset}

    where :math:`y` is the CSDM dependent variable.

    Args:
        float offset: The offset factor. The default value is 0.

    Example
    -------

    >>> from mrsimulator import signal_processing as sp
    >>> operation1 = sp.baseline.ConstantOffset(offset=20)
    """

    offset: float = 0

    def operate(self, data):
        """Applies the operation for which the class is named for.

        Args:
            data: CSDM object
        """
        data += self.offset
        return data



# -*- coding: utf-8 -*-
import os

from mrsimulator import methods as NamedMethods
from mrsimulator import Simulator
from mrsimulator import SpinSystem

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

methods = [val for k, val in NamedMethods.__dict__.items() if isinstance(val, type)]

sys = SpinSystem(sites=[{"isotope": "1H"}])
sp = {"count": 1024, "spectral_width": 100}


def test_read_write_methods():
    for item in methods:
        kwargs = {"channels": ["93Nb"], "spectral_dimensions": [sp] * item.ndim}

        if item.__name__ == "SSB2D":
            kwargs["rotor_frequency"] = 1e3

        fn1 = item(**kwargs)
        fn2 = item.parse_dict_with_units(fn1.json())
        assert fn1 == fn2, f"Error with {item} parse."

        sim = Simulator(spin_systems=[sys], methods=[fn1])
        sim.run()
        sim.save("test.mrsim")

        sim2 = Simulator.load("test.mrsim")
        for mth in sim2.methods:
            mth.simulation._timestamp = ""
            _ = [item.to("ppm", "nmr_frequency_ratio") for item in mth.simulation.x]

        assert sim == sim2, f"Error with {item} parse from Simulator.load()."

    os.remove("test.mrsim")



cimport base_model as clib
from libcpp cimport bool as bool_t
from numpy cimport ndarray
import numpy as np
import cython

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

clib.generate_table()

@cython.profile(False)
@cython.boundscheck(False)
@cython.wraparound(False)
def one_d_spectrum(method,
       list spin_systems,
       int verbose=0,
       unsigned int number_of_sidebands=90,
       unsigned int integration_density=72,
       unsigned int decompose_spectrum=0,
       unsigned int integration_volume=1,
       bool_t interpolation=True):
    """

    :ivar verbose:
        The allowed values are 0, 1, and 11. When the value is 1, the output is
        printed on the screen. When the value is 11, in addition to the output
        from 1, execution time is also printed on the screen.
        The default value is 0.
    :ivar number_of_sidebands:
        The value is an integer which corresponds to the number of sidebands
        simulated in the spectrum. The default value is 90. Note, when the
        sample spinning frequency is low, computation of more sidebands may be
        required for an acceptable result. The user is advised to ensure that
        enough sidebands are requested for computation.
    :ivar integration_density:
        The value is an integer which represents the frequency of class I
        geodesic polyhedra. These polyhedra are used in calculating the
        spherical average. Presently we only use octahedral as the frequency1
        polyhedra. As the frequency of the geodesic polyhedron increases, the
        polyhedra approach a sphere geometry. A higher frequency will result in a
        better powder averaging. The default value is 72.
        Read more on the `Geodesic polyhedron <https://en.wikipedia.org/wiki/Geodesic_polyhedron>`_.
    :ivar decompose_spectrum:
        An unsigned integer. When value is 0, the spectrum is a sum of spectrum from all
        spin systems. If value is 1, spectrum from individual spin systems is stored
        separately.
    """

# observed spin _______________________________________________________
    channel = method.channels[0].symbol
    # spin quantum number of the observed spin
    cdef double spin_quantum_number = method.channels[0].spin

    # gyromagnetic ratio
    cdef gyromagnetic_ratio = method.channels[0].gyromagnetic_ratio
    cdef double factor = 1.0
    if gyromagnetic_ratio > 0.0:
        factor = -1.0

    # if verbose in [1, 11]:
    #     print(f'Simulating {isotope} (I={spin_quantum_number})')
    #     print(f'Larmor frequency (ω0 = - γ B0) = {larmor_frequency/1.0e6} MHz')
    #     print((f'Recording {isotope} spectrum with {number_of_points} '
    #             f'points over {spectral_width} Hz bandwidth'))
    #     print((f"and a reference offset of {dimension['reference_offset']} Hz."))

    # transitions of the observed spin
    cdef int transition_increment
    cdef ndarray[float, ndim=1] transition_pathway_c
    cdef int number_of_transitions
    # transition_pathway_c = np.asarray([-0.5, 0.5]).ravel()
    # number_of_transitions = int(transition_pathway_c.size/2)
    # else:
    #     energy_level_count = int(2*spin_quantum_number+1)
    #     number_of_transitions = energy_level_count-1
    #     energy_states = np.arange(energy_level_count) - spin_quantum_number
    #     transitions = [ [energy_states[i], energy_states[i+1]] for i in range(number_of_transitions)]
    #     transition_pathway_c = np.asarray(transitions).ravel()

    cdef bool_t allow_fourth_rank = 0
    if spin_quantum_number > 0.5:
        allow_fourth_rank = 1

# create averaging scheme _____________________________________________________
    cdef clib.MRS_averaging_scheme *the_averaging_scheme
    the_averaging_scheme = clib.MRS_create_averaging_scheme(
        integration_density=integration_density, allow_fourth_rank=allow_fourth_rank,
        integration_volume=integration_volume
    )

# create spectral dimensions _______________________________________________
    cdef int n_dimension = len(method.spectral_dimensions)
    # if n_sequence > 1:
    #     number_of_sidebands = 1

    max_n_sidebands = number_of_sidebands

    total_n_points = 1
    cdef ndarray[int] n_event
    cdef ndarray[double] magnetic_flux_density_in_T, frac
    cdef ndarray[double] srfiH
    cdef ndarray[double] rair
    cdef ndarray[int] cnt
    cdef ndarray[double] coord_off
    cdef ndarray[double] incre
    freq_contrib = np.asarray([])

    fr = []
    Bo = []
    vr = []
    th = []
    event_i = []
    count = []
    increment = []
    coordinates_offset = []

    prev_n_sidebands = 0
    for i, dim in enumerate(method.spectral_dimensions):
        for event in dim.events:
            freq_contrib = np.append(freq_contrib, event._freq_contrib_flags())
            if event.rotor_frequency < 1.0e-3:
                rotor_frequency_in_Hz = 1.0e9
                rotor_angle_in_rad = 0.0
                number_of_sidebands = 1
                if prev_n_sidebands == 0: prev_n_sidebands = 1
            else:
                rotor_frequency_in_Hz = event.rotor_frequency
                rotor_angle_in_rad = event.rotor_angle
                if prev_n_sidebands == 0: prev_n_sidebands = number_of_sidebands

            if prev_n_sidebands != number_of_sidebands:
                raise ValueError(
                    (
                        'The library does not support spectral dimensions containing '
                        'both zero and non-zero rotor frequencies. Consider using a '
                        'smaller value instead of zero.'
                    )
                )

            fr.append(event.fraction) # fraction
            Bo.append(event.magnetic_flux_density)  # in T
            vr.append(rotor_frequency_in_Hz) # in Hz
            th.append(rotor_angle_in_rad) # in rad

        total_n_points *= dim.count

        count.append(dim.count)
        offset = dim.spectral_width / 2.0
        coordinates_offset.append(-dim.reference_offset * factor - offset)
        increment.append(dim.spectral_width / dim.count)
        event_i.append(len(dim.events))

        if dim.origin_offset is None:
            dim.origin_offset = np.abs(Bo[0] * gyromagnetic_ratio * 1e6)

    frac = np.asarray(fr, dtype=np.float64)
    magnetic_flux_density_in_T = np.asarray(Bo, dtype=np.float64)
    srfiH = np.asarray(vr, dtype=np.float64)
    rair = np.asarray(th, dtype=np.float64)
    cnt = np.asarray(count, dtype=np.int32)
    incre = np.asarray(increment, dtype=np.float64)
    coord_off = np.asarray(coordinates_offset, dtype=np.float64)
    n_event = np.asarray(event_i, dtype=np.int32)

    # create spectral_dimensions
    dimensions = clib.MRS_create_dimensions(the_averaging_scheme, &cnt[0],
        &coord_off[0], &incre[0], &frac[0], &magnetic_flux_density_in_T[0],
        &srfiH[0], &rair[0], &n_event[0], n_dimension, number_of_sidebands)

# normalization factor for the spectrum
    norm = np.prod(incre)

# create fftw scheme __________________________________________________________
    cdef clib.MRS_fftw_scheme *the_fftw_scheme
    the_fftw_scheme = clib.create_fftw_scheme(the_averaging_scheme.total_orientations, number_of_sidebands)

# _____________________________________________________________________________

# frequency contrib
    cdef ndarray[bool_t] freq_contrib_c = np.asarray(freq_contrib, dtype=bool)

# affine transformation
    cdef ndarray[double] affine_matrix_c
    if method.affine_matrix is None:
        affine_matrix_c = np.asarray([1, 0, 0, 1], dtype=np.float64)
    else:
        increment_fraction = [incre/item for item in incre]
        matrix = method.affine_matrix.ravel() * np.asarray(increment_fraction).ravel()
        affine_matrix_c = np.asarray(matrix, dtype=np.float64)
        if affine_matrix_c[2] != 0:
            affine_matrix_c[2] /= affine_matrix_c[0]
            affine_matrix_c[3] -=  affine_matrix_c[1]*affine_matrix_c[2]

    # B0 = dimension.magnetic_flux_density

    # if verbose in [1, 11]:
    #     text = "`one_d_spectrum` method simulation parameters."
    #     len_ = len(text)
    #     print(text)
    #     print(f"{'-'*(len_-1)}")
    #     print (f'Macroscopic magnetic flux density (B0) = {B0} T')
    #     print (f'Sample rotation angle is (θ) = {rotor_angle_in_rad} rad')
    #     print (f'Sample rotation frequency (𝜈r) = {rotor_frequency_in_Hz} Hz')

# sites _______________________________________________________________________________
    p_isotopes = None

    cdef int number_of_sites, number_of_couplings, p_number_of_sites=0
    cdef ndarray[int] spin_index_ij
    cdef ndarray[float] spin_i
    cdef ndarray[double] gyromagnetic_ratio_i

    # CSA
    cdef ndarray[double] iso_n
    cdef ndarray[double] zeta_n
    cdef ndarray[double] eta_n
    cdef ndarray[double] ori_n

    # quad
    cdef ndarray[double] Cq_e
    cdef ndarray[double] eta_e
    cdef ndarray[double] ori_e

    # J-coupling
    cdef ndarray[double] iso_j
    cdef ndarray[double] zeta_j
    cdef ndarray[double] eta_j
    cdef ndarray[double] ori_j

    # quad
    cdef ndarray[double] D_d
    cdef ndarray[double] eta_d
    cdef ndarray[double] ori_d



    cdef int trans__, pathway_increment, pathway_count, transition_count_per_pathway
    cdef ndarray[double, ndim=1] amp
    amp1 = np.zeros(total_n_points, dtype=np.float64)
    amp_individual = []

    cdef clib.site_struct sites_c
    cdef clib.coupling_struct couplings_c

    index_ = []

    # -------------------------------------------------------------------------
    # sample __________________________________________________________________
    for index, spin_sys in enumerate(spin_systems):
        abundance = spin_sys.abundance
        isotopes = [site.isotope.symbol for site in spin_sys.sites]
        if channel not in isotopes:
            if decompose_spectrum == 1:
                amp_individual.append(np.zeros(method.shape()))
            continue

        # sub_sites = [site for site in spin_sys.sites if site.isotope.symbol == isotope]
        index_.append(index)
        number_of_sites = len(spin_sys.sites)

        # ------------------------------------------------------------------------
        #                          Site specification
        # ------------------------------------------------------------------------
        # CSA
        spin_i = np.empty(number_of_sites, dtype=np.float32)
        gyromagnetic_ratio_i = np.empty(number_of_sites, dtype=np.float64)

        iso_n = np.zeros(number_of_sites, dtype=np.float64)
        zeta_n = np.zeros(number_of_sites, dtype=np.float64)
        eta_n = np.zeros(number_of_sites, dtype=np.float64)
        ori_n = np.zeros(3*number_of_sites, dtype=np.float64)

        # Quad
        Cq_e = np.zeros(number_of_sites, dtype=np.float64)
        eta_e = np.zeros(number_of_sites, dtype=np.float64)
        ori_e = np.zeros(3*number_of_sites, dtype=np.float64)

        # Extract and assign site information from Site objects to C structure
        # ---------------------------------------------------------------------
        for i in range(number_of_sites):
            site = spin_sys.sites[i]
            spin_i[i] = site.isotope.spin
            gyromagnetic_ratio_i[i] = site.isotope.gyromagnetic_ratio
            i3 = 3*i

            # CSA tensor
            if site.isotropic_chemical_shift is not None:
                iso_n[i] = site.isotropic_chemical_shift

            shielding = site.shielding_symmetric
            if shielding is not None:
                if shielding.zeta is not None:
                    zeta_n[i] = shielding.zeta
                if shielding.eta is not None:
                    eta_n[i] = shielding.eta
                if shielding.alpha is not None:
                    ori_n[i3] = shielding.alpha
                if shielding.beta is not None:
                    ori_n[i3+1] = shielding.beta
                if shielding.gamma is not None:
                    ori_n[i3+2] = shielding.gamma

            # if verbose in [1, 11]:
            #     text = ((
            #         f"\n{isotope} site {i} from spin system {index} "
            #         f"@ {abundance}% abundance"
            #     ))
            #     len_ = len(text)
            #     print(text)
            #     print(f"{'-'*(len_-1)}")
            #     print(f'Isotropic chemical shift (δ) = {str(1e6*iso/larmor_frequency)} ppm')
            #     print(f'Shielding anisotropy (ζ) = {str(1e6*zeta/larmor_frequency)} ppm')
            #     print(f'Shielding asymmetry (η) = {eta}')
            #     print(f'Shielding orientation = [alpha = {alpha}, beta = {beta}, gamma = {gamma}]')

            # quad tensor
            if spin_quantum_number > 0.5:
                quad = site.quadrupolar
                if quad is not None:
                    if quad.Cq is not None:
                        Cq_e[i] = quad.Cq
                    if quad.eta is not None:
                        eta_e[i] = quad.eta
                    if quad.alpha is not None:
                        ori_e[i3] = quad.alpha
                    if quad.beta is not None:
                        ori_e[i3+1] = quad.beta
                    if quad.gamma is not None:
                        ori_e[i3+2] = quad.gamma

                # if verbose in [1, 11]:
                #     print(f'Quadrupolar coupling constant (Cq) = {Cq_e[i]/1e6} MHz')
                #     print(f'Quadrupolar asymmetry (η) = {eta}')
                #     print(f'Quadrupolar orientation = [alpha = {alpha}, beta = {beta}, gamma = {gamma}]')

        # sites packed as c struct
        sites_c.number_of_sites = number_of_sites
        sites_c.spin = &spin_i[0]
        sites_c.gyromagnetic_ratio = &gyromagnetic_ratio_i[0]

        sites_c.isotropic_chemical_shift_in_ppm = &iso_n[0]
        sites_c.shielding_symmetric_zeta_in_ppm = &zeta_n[0]
        sites_c.shielding_symmetric_eta = &eta_n[0]
        sites_c.shielding_orientation = &ori_n[0]

        sites_c.quadrupolar_Cq_in_Hz = &Cq_e[0]
        sites_c.quadrupolar_eta = &eta_e[0]
        sites_c.quadrupolar_orientation = &ori_e[0]
        # ------------------------------------------------------------------------
        #                           Coupling specification
        # ------------------------------------------------------------------------
        # J-coupling
        couplings_c.number_of_couplings = 0
        if spin_sys.couplings is not None:
            number_of_couplings = len(spin_sys.couplings)
            spin_index_ij = np.zeros(2*number_of_couplings, dtype=np.int32)

            iso_j = np.zeros(number_of_couplings, dtype=np.float64)
            zeta_j = np.zeros(number_of_couplings, dtype=np.float64)
            eta_j = np.zeros(number_of_couplings, dtype=np.float64)
            ori_j = np.zeros(3*number_of_couplings, dtype=np.float64)

            # Dipolar
            D_d = np.zeros(number_of_couplings, dtype=np.float64)
            eta_d = np.zeros(number_of_couplings, dtype=np.float64)
            ori_d = np.zeros(3*number_of_couplings, dtype=np.float64)

            # Extract and assign coupling information from Site objects to C structure
            for i in range(number_of_couplings):
                coupling = spin_sys.couplings[i]
                spin_index_ij[2*i: 2*i+2] = coupling.site_index
                i3 = 3*i

                # J tensor
                if coupling.isotropic_j is not None:
                    iso_j[i] = coupling.isotropic_j

                J_sym = coupling.j_symmetric
                if J_sym is not None:
                    if J_sym.zeta is not None:
                        zeta_j[i] = J_sym.zeta
                    if J_sym.eta is not None:
                        eta_j[i] = J_sym.eta
                    if J_sym.alpha is not None:
                        ori_j[i3] = J_sym.alpha
                    if J_sym.beta is not None:
                        ori_j[i3+1] = J_sym.beta
                    if J_sym.gamma is not None:
                        ori_j[i3+2] = J_sym.gamma

                # dipolar tensor
                dipolar = coupling.dipolar
                if dipolar is not None:
                    if dipolar.D is not None:
                        D_d[i] = dipolar.D
                    if dipolar.eta is not None:
                        eta_d[i] = dipolar.eta
                    if dipolar.alpha is not None:
                        ori_d[i3] = dipolar.alpha
                    if dipolar.beta is not None:
                        ori_d[i3+1] = dipolar.beta
                    if dipolar.gamma is not None:
                        ori_d[i3+2] = dipolar.gamma

            # if verbose in [1, 11]:
            #     print(f'N couplings = {number_of_couplings}')
            #     print(f'site index J = {spin_index_ij}')
            #     print(f'Isotropic J = {iso_j} Hz')
            #     print(f'J anisotropy = {zeta_j} Hz')
            #     print(f'J asymmetry = {eta_j}')
            #     print(f'J orientation = {ori_j}')

            #     print(f'Dipolar coupling constant = {D_d} Hz')
            #     print(f'Dipolar asymmetry = {eta_d}')
            #     print(f'Dipolar orientation = {ori_d}')

            # couplings packed as c struct
            couplings_c.number_of_couplings = number_of_couplings
            couplings_c.site_index = &spin_index_ij[0]

            couplings_c.isotropic_j_in_Hz = &iso_j[0]
            couplings_c.j_symmetric_zeta_in_Hz = &zeta_j[0]
            couplings_c.j_symmetric_eta = &eta_j[0]
            couplings_c.j_orientation = &ori_j[0]

            couplings_c.dipolar_coupling_in_Hz = &D_d[0]
            couplings_c.dipolar_eta = &eta_d[0]
            couplings_c.dipolar_orientation = &ori_d[0]


        # Spectrum amplitude vector -------------------------------------------
        amp = np.zeros(total_n_points)

        # if number_of_sites == 0:
        #     if decompose_spectrum == 1:
        #         amp_individual.append([])
        #     continue

        if number_of_sites != p_number_of_sites and isotopes != p_isotopes:
            transition_pathway = spin_sys.transition_pathways
            if transition_pathway is None:
                transition_pathway = np.asarray(method._get_transition_pathways_np(spin_sys))
                transition_pathway_c = np.asarray(transition_pathway, dtype=np.float32).ravel()
            else:
                transition_pathway = np.asarray(transition_pathway)
                # convert transition objects to list
                lst = [item.tolist() for item in transition_pathway.ravel()]
                transition_pathway_c = np.asarray(lst, dtype=np.float32).ravel()

            pathway_count, transition_count_per_pathway = transition_pathway.shape[:2]
            pathway_increment = 2*number_of_sites*transition_count_per_pathway

            p_number_of_sites = number_of_sites
            p_isotopes = isotopes

        # if spin_sys.transitions is not None:
        #     transition_pathway_c = np.asarray(
        #         spin_sys.transitions, dtype=np.float32
        #     ).ravel()
        # else:
        #     transition_pathway_c = np.asarray([0.5, -0.5], dtype=np.float32)

        # the number 2 is because of single site transition [mi, mf]
        # it dose not work for coupled sites.
        # transition_increment = 2*number_of_sites
        # number_of_transitions = int((transition_pathway_c.size)/transition_increment)

        for trans__ in range(pathway_count):
            clib.__mrsimulator_core(
                # spectrum information and related amplitude
                &amp[0],
                &sites_c,
                &couplings_c,
                &transition_pathway_c[pathway_increment*trans__],
                n_dimension,          # The total number of spectroscopic dimensions.
                dimensions,           # Pointer to MRS_dimension structure
                the_fftw_scheme,      # Pointer to the fftw scheme.
                the_averaging_scheme, # Pointer to the powder averaging scheme.
                interpolation,
                &freq_contrib_c[0],
                &affine_matrix_c[0],
                )

        temp = amp*abundance/norm

        if decompose_spectrum == 1:
            amp_individual.append(temp.reshape(method.shape()))
        else:
            amp1 += temp
        # else:
        #     if decompose_spectrum == 1:
        #         amp_individual.append([])

    # reverse the spectrum if gyromagnetic ratio is positive.
    if decompose_spectrum == 1 and len(amp_individual) != 0:
        if gyromagnetic_ratio < 0:
            amp1 = [np.fft.fftn(np.fft.ifftn(item).conj()).real for item in amp_individual]
        else:
            amp1 = amp_individual
    else:
        amp1.shape = method.shape()
        if gyromagnetic_ratio < 0:
            amp1 = np.fft.fftn(np.fft.ifftn(amp1).conj()).real

    clib.MRS_free_dimension(dimensions, n_dimension)
    clib.MRS_free_averaging_scheme(the_averaging_scheme)
    clib.MRS_free_fftw_scheme(the_fftw_scheme)
    return amp1


@cython.profile(False)
@cython.boundscheck(False)
@cython.wraparound(False)
def get_zeeman_states(sys):
    cdef int i, j, n_site = len(sys.sites)

    two_Ip1 = [int(2 * site.isotope.spin + 1) for site in sys.sites]
    spin_quantum_numbers = [
        np.arange(two_Ip1[i]) - site.isotope.spin for i, site in enumerate(sys.sites)
    ]

    lst = []
    for j in range(n_site):
        k = 1
        for i in range(n_site):
            if i == j:
                k = np.kron(k, spin_quantum_numbers[i])
            else:
                k = np.kron(k, np.ones(two_Ip1[i]))
        lst.append(k)
    return np.asarray(lst).T



# -*- coding: utf-8 -*-
# import pytest
# from mrsimulator.method.transition_query import TransitionQuery
# from mrsimulator.methods import Cosy
# from mrsimulator.methods import Inadequate
#
# __author__ = "Deepansh J. Srivastava"
# __email__ = "srivastava.89@osu.edu"
#
# methods = [Cosy, Inadequate]
# names = ["Cosy", "Inadequate"]
#
#
# def test_coorelation_rotor_freq():
#     def error(name):
#         return f"`rotor_frequency` value cannot be modified for {name} method."
#
#     for name, method in zip(names, methods):
#         e = error(name)
#         with pytest.raises(ValueError, match=f".*{e}.*"):
#             method(rotor_frequency=10, spectral_dimensions=[{}, {}])
#
#
# def test_coorelation_spectral_dimension_count():
#     e = "Method requires exactly 2 spectral dimensions, given 1."
#     for _, method in zip(names, methods):
#         with pytest.raises(ValueError, match=f".*{e}.*"):
#             method(spectral_dimensions=[{}])
#
#
# def test_coorelation_setting_transition_query():
#     def error(name):
#         return f"`transition_query` value cannot be modified for {name} method."
#
#     for name, method in zip(names, methods):
#         e = error(name)
#         with pytest.raises(ValueError, match=f".*{e}.*"):
#             method(
#                 spectral_dimensions=[
#                     {"events": [{"transition_query": {"P": [-1]}}]},
#                     {},
#                 ],
#             )
#
#
# def test_Cosy_general():
#     """Inner satellite-transition variable-angle spinning method"""
#     mth = Cosy(
#         channels=["13C"],
#         magnetic_flux_density=9.4,  # in T
#         spectral_dimensions=[
#             {
#                 "count": 1024,
#                 "spectral_width": 5e4,  # in Hz
#                 "reference_offset": 0,  # in Hz
#             },
#             {
#                 "count": 1024,
#                 "spectral_width": 5e4,  # in Hz
#                 "reference_offset": 0,  # in Hz
#             },
#         ],
#     )
#     assert mth.name == "Cosy"
#
#     des = "Simulate an infinite spinning COrrelation SpectroscopY spectrum."
#     assert mth.description == des
#     assert mth.spectral_dimensions[0].events[0].transition_query == TransitionQuery(
#         P={"channel-1": [[-1]]}
#     )
#     assert mth.spectral_dimensions[1].events[0].transition_query == TransitionQuery(
#         P={"channel-1": [[-1]]}
#     )
#     assert Cosy.parse_dict_with_units(mth.json()) == mth
#
#     serialize = mth.json()
#     assert serialize == {
#         "channels": ["13C"],
#         "description": des,
#         "magnetic_flux_density": "9.4 T",
#         "name": "Cosy",
#         "rotor_angle": "0.955316618 rad",
#         "rotor_frequency": "1000000000000.0 Hz",
#         "spectral_dimensions": [
#             {
#                 "count": 1024,
#                 "reference_offset": "0.0 Hz",
#                 "spectral_width": "50000.0 Hz",
#             },
#             {
#                 "count": 1024,
#                 "reference_offset": "0.0 Hz",
#                 "spectral_width": "50000.0 Hz",
#             },
#         ],
#     }
#
#
# def test_Inadequate_general():
#     """Second to inner satellite-transition variable-angle spinning method"""
#     mth = Inadequate(
#         channels=["17O"],
#         magnetic_flux_density=9.4,  # in T
#         spectral_dimensions=[
#             {
#                 "count": 1024,
#                 "spectral_width": 5e4,  # in Hz
#                 "reference_offset": 0,  # in Hz
#             },
#             {
#                 "count": 1024,
#                 "spectral_width": 5e4,  # in Hz
#                 "reference_offset": 0,  # in Hz
#             },
#         ],
#     )
#     assert mth.name == "Inadequate"
#
#     des = (
#         "Simulate an infinite spinning Incredible Natural Abundance DoublE QUAntum "
#         "Transfer Experiment spectrum."
#     )
#     assert mth.description == des
#     assert mth.spectral_dimensions[0].events[0].transition_query == TransitionQuery(
#         P={"channel-1": [[-1, -1]]}
#     )
#     assert mth.spectral_dimensions[1].events[0].transition_query == TransitionQuery(
#         P={"channel-1": [[-1]]}
#     )
#     assert Inadequate.parse_dict_with_units(mth.json()) == mth
#
#     serialize = mth.json()
#     assert serialize == {
#         "channels": ["17O"],
#         "description": des,
#         "magnetic_flux_density": "9.4 T",
#         "name": "Inadequate",
#         "rotor_angle": "0.955316618 rad",
#         "rotor_frequency": "1000000000000.0 Hz",
#         "spectral_dimensions": [
#             {
#                 "count": 1024,
#                 "reference_offset": "0.0 Hz",
#                 "spectral_width": "50000.0 Hz",
#             },
#             {
#                 "count": 1024,
#                 "reference_offset": "0.0 Hz",
#                 "spectral_width": "50000.0 Hz",
#             },
#         ],
#     }



# -*- coding: utf-8 -*-
import numpy as np
import pytest
from mrsimulator.method.transition_query import TransitionQuery
from mrsimulator.methods import ST1_VAS
from mrsimulator.methods import ST2_VAS

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

methods = [ST1_VAS, ST2_VAS]
names = ["ST1_VAS", "ST2_VAS"]


def test_ST_VAS_rotor_freq():
    def error(name):
        return f"`rotor_frequency` value cannot be modified for {name} method."

    for name, method in zip(names, methods):
        e = error(name)
        with pytest.raises(ValueError, match=f".*{e}.*"):
            method(rotor_frequency=10, spectral_dimensions=[{}, {}])


def test_ST_VAS_spectral_dimension_count():
    e = "Method requires exactly 2 spectral dimensions, given 1."
    for _, method in zip(names, methods):
        with pytest.raises(ValueError, match=f".*{e}.*"):
            method(spectral_dimensions=[{}])


def test_ST_VAS_setting_transition_query():
    def error(name):
        return f"`transition_query` value cannot be modified for {name} method."

    for name, method in zip(names, methods):
        e = error(name)
        with pytest.raises(ValueError, match=f".*{e}.*"):
            method(
                spectral_dimensions=[
                    {"events": [{"transition_query": {"P": [-1]}}]},
                    {},
                ],
            )


def test_ST_VAS_affine():
    sites = ["87Rb", "27Al"]
    spins = [1.5, 2.5]
    k_ST_MAS = {
        3: {1.5: 24 / 27, 2.5: 21 / 72, 3.5: 84 / 135, 4.5: 165 / 216},
        5: {2.5: 132 / 72, 3.5: 69 / 135, 4.5: 12 / 216},
        7: {3.5: 324 / 135, 4.5: 243 / 216},
        9: {4.5: 600 / 216},
    }
    for j, method in enumerate(methods):
        for i, isotope in zip(spins[j:], sites[j:]):
            meth = method(channels=[isotope])
            k = k_ST_MAS[3 + 2 * j][i]
            assert meth.spectral_dimensions[0].events[0].fraction == 1
            assert meth.spectral_dimensions[1].events[0].fraction == 1
            assert np.allclose(meth.affine_matrix, [1 / (1 + k), k / (1 + k), 0, 1])


def test_ST1_VAS_general():
    """Inner satellite-transition variable-angle spinning method"""
    mth = ST1_VAS(
        channels=["87Rb"],
        magnetic_flux_density=11.7,  # in T
        spectral_dimensions=[
            {"count": 1024, "spectral_width": 3e4},
            {"count": 1024, "spectral_width": 2e4},
        ],
    )
    assert mth.name == "ST1_VAS"

    des = (
        "Simulate a 1.5 -> 0.5 and -0.5 -> -1.5 satellite-transition variable-angle "
        "spinning spectrum."
    )
    assert mth.description == des
    assert mth.spectral_dimensions[0].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-1]]}, D={"channel-1": [[2], [-2]]}
    )
    assert mth.spectral_dimensions[1].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-1]]}, D={"channel-1": [[0]]}
    )
    assert ST1_VAS.parse_dict_with_units(mth.json()) == mth

    assert np.allclose(mth.affine_matrix, [0.52941176, 0.47058824, 0.0, 1.0])

    serialize = mth.json()
    _ = serialize.pop("affine_matrix")

    assert serialize == {
        "channels": ["87Rb"],
        "description": des,
        "magnetic_flux_density": "11.7 T",
        "name": "ST1_VAS",
        "rotor_angle": "0.955316618 rad",
        "rotor_frequency": "1000000000000.0 Hz",
        "spectral_dimensions": [
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "30000.0 Hz",
            },
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "20000.0 Hz",
            },
        ],
    }


def test_ST2_VAS_general():
    """Second to inner satellite-transition variable-angle spinning method"""
    mth = ST2_VAS(
        channels=["17O"],
        magnetic_flux_density=9.4,  # in T
        spectral_dimensions=[
            {"count": 1024, "spectral_width": 5e4},
            {"count": 1024, "spectral_width": 5e4},
        ],
    )
    assert mth.name == "ST2_VAS"

    des = (
        "Simulate a 2.5 -> 1.5 and -1.5 -> -2.5 satellite-transition variable-angle "
        "spinning spectrum."
    )
    assert mth.description == des
    assert mth.spectral_dimensions[0].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-1]]}, D={"channel-1": [[4], [-4]]}
    )
    assert mth.spectral_dimensions[1].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-1]]}, D={"channel-1": [[0]]}
    )
    assert ST2_VAS.parse_dict_with_units(mth.json()) == mth

    assert np.allclose(mth.affine_matrix, [0.35294118, 0.64705882, 0.0, 1.0])

    serialize = mth.json()
    _ = serialize.pop("affine_matrix")

    assert serialize == {
        "channels": ["17O"],
        "description": des,
        "magnetic_flux_density": "9.4 T",
        "name": "ST2_VAS",
        "rotor_angle": "0.955316618 rad",
        "rotor_frequency": "1000000000000.0 Hz",
        "spectral_dimensions": [
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "50000.0 Hz",
            },
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "50000.0 Hz",
            },
        ],
    }



# -*- coding: utf-8 -*-
import numpy as np
from csdmpy.units import string_to_quantity


__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"

CONST = string_to_quantity("1")


def _get_broadcast_shape(array, dim, ndim):
    """Return the broadcast shape of a vector `array` at dimension `dim` for `ndim`
    total dimensions."""
    none = [np.newaxis for _ in range(ndim + 1)]
    dim = [dim] if isinstance(dim, int) else dim
    for dim_ in dim:
        none[-dim_ - 1] = slice(None, None, None)
    return array[tuple(none)]


def _str_to_quantity(v, values, prop_name):
    if isinstance(v, str):
        quantity = string_to_quantity(v)
        values["property_units"] = {prop_name: quantity.unit}
        return quantity.value
    if isinstance(v, float):
        return v



# -*- coding: utf-8 -*-
from enum import Enum

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

freq_list_all = [
    "Shielding1_0",
    "Shielding1_2",
    "Quad1_2",
    "Quad2_0",
    "Quad2_2",
    "Quad2_4",
]


class FrequencyEnum(str, Enum):
    """Enumeration for selecting specific frequency contributions. The enumerations
    are:

    Attributes
    ----------

    Shielding1_0:
        Selects first-order and zeroth-rank nuclear shielding frequency contributions.

    Shielding1_2:
        Selects first-order and second-rank nuclear shielding frequency contributions.

    Quad1_2:
        Selects first-order and second-rank quadrupolar frequency contributions.

    Quad2_0:
        Selects second-order and zeroth-rank quadrupolar frequency contributions.

    Quad2_2:
        Selects second-order and second-rank quadrupolar frequency contributions.

    Quad2_4:
        Selects second-order and fourth-rank quadrupolar frequency contributions.
    """

    # Shielding1:
    #     Selects first-order nuclear shielding frequency contributions. Equivalent to
    #     ["Shielding1_0", "Shielding1_2"]
    # Quad1:
    #     Selects first-order quadrupolar frequency contributions. Equivalent to
    #     ["Quad1_2"]
    # Quad2:
    #     Selects second-order quadrupolar frequency contributions. Equivalent to
    #     ["Quad2_0", "Quad2_2", "Quad2_4"]

    # Shielding1: str = "Shielding1"
    Shielding1_0: str = freq_list_all[0]
    Shielding1_2: str = freq_list_all[1]

    # Quad1: str = "Quad1"
    Quad1_2: str = freq_list_all[2]

    # Quad2: str = "Quad2"
    Quad2_0: str = freq_list_all[3]
    Quad2_2: str = freq_list_all[4]
    Quad2_4: str = freq_list_all[5]

    # @validator("Quad1", pre=True, always=True)
    # def validate_Quad1(cls, v, *, values, **kwargs):
    #     values["Quad1_2"] = "Quad1_2"
    #     return None

    # @validator("Quad2", pre=True, always=True)
    # def validate_Quad2(cls, v, *, values, **kwargs):
    #     for item in ["Quad2_0", "Quad2_2", "Quad2_4"]:
    #         values[item] = item
    #     return None

    def json(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object."""
        return self.value

    def index(self) -> int:
        """Get the index of enumeration relative to freq_list_all."""
        return freq_list_all.index(self.value)


freq_default = [
    "Shielding1_0",
    "Shielding1_2",
    "Quad1_2",
    "Quad2_0",
    "Quad2_2",
    "Quad2_4",
]
default_freq_contrib = [FrequencyEnum(item) for item in freq_default]



# -*- coding: utf-8 -*-
"""Apodization test"""
import csdmpy as cp
import numpy as np
from mrsimulator import signal_processing as sp
from mrsimulator import Simulator
from mrsimulator import SpinSystem
from mrsimulator.methods import BlochDecaySpectrum

from .test_signal_processing import setup_read_write

__author__ = "Maxwell C. Venetos"
__email__ = "maxvenetos@gmail.com"

sim = Simulator()
the_site = {"isotope": "1H", "isotropic_chemical_shift": "0 ppm"}
the_spin_system = {"name": "site A", "sites": [the_site], "abundance": "80%"}
spin_system_object = SpinSystem.parse_dict_with_units(the_spin_system)
sim.spin_systems += [spin_system_object, spin_system_object, spin_system_object]
sim.config.decompose_spectrum = "spin_system"

sim.methods += [
    BlochDecaySpectrum(
        channels=["1H"],
        magnetic_flux_density=9.4,
        spectral_dimensions=[{"count": 65536, "spectral_width": 25000}],
    )
]
sim.run()
freqHz = sim.methods[0].spectral_dimensions[0].coordinates_Hz()


def test_scale():
    PS_0 = [sp.Scale(factor=10)]
    post_sim = sp.SignalProcessor(operations=PS_0)
    data = post_sim.apply_operations(data=sim.methods[0].simulation.copy())
    _, y0, y1, y2 = sim.methods[0].simulation.to_list()
    _, y0_, y1_, y2_ = data.to_list()

    assert y0_.max() / y0.max() == 10, "Scaling failed"
    assert y1_.max() / y1.max() == 10, "Scaling failed"
    assert y2_.max() / y2.max() == 10, "Scaling failed"


def test_Lorentzian():
    PS_1 = [
        sp.IFFT(dim_index=0),
        sp.apodization.Exponential(FWHM="200 Hz", dim_index=0, dv_index=0),
        sp.FFT(dim_index=0),
    ]
    post_sim = sp.SignalProcessor(operations=PS_1)
    data = post_sim.apply_operations(data=sim.methods[0].simulation.copy())
    _, y0, y1, y2 = data.to_list()

    FWHM = 200
    test = (FWHM / 2) / (np.pi * (freqHz ** 2 + (FWHM / 2) ** 2))

    assert np.allclose(y1, y2)
    assert np.all(y0 != y1)
    assert np.allclose(
        test / test.sum(), y0 / y0.sum(), atol=1e-04
    ), "Lorentzian apodization amplitude failed"
    assert np.allclose(
        y1.sum(), y0.sum()
    ), "Area not conserved after Lorentzian apodization"


def test_Gaussian():
    FWHM = 200 * 2.354820045030949
    PS_2 = [
        sp.IFFT(dim_index=0),
        sp.apodization.Gaussian(FWHM=f"{FWHM} Hz", dim_index=0, dv_index=[0, 1]),
        sp.FFT(dim_index=0),
    ]

    PS_3 = [
        sp.IFFT(dim_index=0),
        sp.apodization.Gaussian(FWHM=f"{FWHM} Hz", dim_index=0, dv_index=None),
        sp.FFT(dim_index=0),
    ]

    post_sim = sp.SignalProcessor(operations=PS_2)
    data = post_sim.apply_operations(data=sim.methods[0].simulation.copy())
    _, y0, y1, _ = data.to_list()

    sigma = 200
    test = (1 / (sigma * np.sqrt(2 * np.pi))) * np.exp(-((freqHz / sigma) ** 2) / 2)

    assert np.allclose(y0, y1), "Gaussian apodization on two dv are not equal."

    assert np.allclose(
        test / test.sum(), y0 / y0.sum(), atol=1e-04
    ), "Gaussian apodization amplitude failed"

    # test None for dv_index
    post_sim = sp.SignalProcessor(operations=PS_3)
    data = post_sim.apply_operations(data=sim.methods[0].simulation.copy())
    _, y0, y1, y2 = data.to_list()

    assert np.allclose(y0, y1), "Gaussian apodization on dv at 0 and 1 are unequal."
    assert np.allclose(y0, y2), "Gaussian apodization on dv at 0 and 2 are unequal."
    assert np.allclose(
        test / test.sum(), y0 / y0.sum(), atol=1e-04
    ), "Gaussian apodization amplitude failed"


def test_scale_class():
    # direct initialization
    a = sp.Scale(factor=200)

    assert a.factor == 200
    assert a.property_units == {}

    # class to dict with units
    dict_ = a.json()

    assert dict_ == {
        "function": "Scale",
        "factor": 200.0,
    }

    # read from dictionary
    b = sp.Scale.parse_dict_with_units(dict_)

    assert a == b


def test_Exponential_class():
    # direct initialization
    a = sp.apodization.Exponential(FWHM="200 s", dim_index=0, dv_index=0)

    assert a.FWHM == 200
    assert a.property_units == {"FWHM": "s"}
    assert a.dim_index == 0
    assert a.dv_index == 0

    py_dict = {
        "function": "apodization",
        "type": "Exponential",
        "FWHM": "200.0 s",
        "dim_index": 0,
        "dv_index": 0,
    }
    setup_read_write(a, py_dict, sp.apodization.Exponential)


def test_Gaussian_class():
    # direct initialization
    a = sp.apodization.Gaussian(FWHM="200 km/s", dim_index=0, dv_index=0)

    assert a.FWHM == 200
    assert a.property_units == {"FWHM": "km / s"}
    assert a.dim_index == 0
    assert a.dv_index == 0

    py_dict = {
        "function": "apodization",
        "type": "Gaussian",
        "FWHM": "200.0 km / s",
        "dim_index": 0,
        "dv_index": 0,
    }
    setup_read_write(a, py_dict, sp.apodization.Gaussian)


def test_2D_area():
    data = np.zeros((256, 128), dtype=float)
    data[128, 64] = 1.0
    csdm_obj = cp.as_csdm(data)

    # test00
    PS = [
        sp.IFFT(dim_index=(0, 1)),
        sp.apodization.Gaussian(FWHM="35", dim_index=0),
        sp.apodization.Gaussian(FWHM="55", dim_index=1),
        sp.FFT(dim_index=(0, 1)),
    ]

    post_sim = sp.SignalProcessor(operations=PS)
    data_new = post_sim.apply_operations(data=csdm_obj.copy())
    _, __, y1 = data_new.to_list()

    assert np.allclose(y1.sum(), data.sum())

    # test01
    PS = [
        sp.IFFT(dim_index=(0, 1)),
        sp.apodization.Gaussian(FWHM="35", dim_index=0),
        sp.apodization.Exponential(FWHM="55", dim_index=1),
        sp.FFT(dim_index=(0, 1)),
    ]

    post_sim = sp.SignalProcessor(operations=PS)
    data_new = post_sim.apply_operations(data=csdm_obj.copy())
    _, __, y1 = data_new.to_list()

    assert np.allclose(y1.sum(), data.sum())



# -*- coding: utf-8 -*-
#
#  nmr_method.pxd
#
#  @copyright Deepansh J. Srivastava, 2019-2021.
#  Created by Deepansh J. Srivastava.
#  Contact email = srivastava.89@osu.edu
#
from libcpp cimport bool as bool_t

cdef extern from "tables/trig.h":
    void generate_table()

cdef extern from "angular_momentum.h":
    void wigner_d_matrices_from_exp_I_beta(int l, int n, bool_t half,
                                void *exp_I_beta, double *wigner)


cdef extern from "schemes.h":
    ctypedef struct MRS_averaging_scheme:
        unsigned int total_orientations

    ctypedef struct MRS_fftw_scheme:
        pass

    MRS_averaging_scheme *MRS_create_averaging_scheme(
                            unsigned int integration_density,
                            bool_t allow_fourth_rank,
                            unsigned int integration_volume)
    void MRS_free_averaging_scheme(MRS_averaging_scheme *scheme)
    MRS_fftw_scheme *create_fftw_scheme(unsigned int total_orientations,
                                    unsigned int number_of_sidebands)
    void MRS_free_fftw_scheme(MRS_fftw_scheme *fftw_scheme)


cdef extern from "mrsimulator.h":
    ctypedef struct MRS_plan:
        MRS_averaging_scheme *averaging_scheme
        unsigned int number_of_sidebands
        double rotor_frequency_in_Hz
        double rotor_angle_in_rad
        # double complex *vector

    MRS_plan *MRS_create_plan(MRS_averaging_scheme *scheme, unsigned int number_of_sidebands,
                          double rotor_frequency_in_Hz,
                          double rotor_angle_in_rad, double increment,
                          bool_t allow_fourth_rank)
    void MRS_free_plan(MRS_plan *plan)
    void MRS_get_amplitudes_from_plan(MRS_plan *plan, bool_t refresh)
    void MRS_get_frequencies_from_plan(MRS_plan *plan, double R0, double complex *R2,
                                  double complex *R4, bool_t refresh)

cdef extern from "object_struct.h":
    ctypedef struct site_struct:
        int number_of_sites                     # Number of sites
        float *spin                             # The spin quantum number
        double *gyromagnetic_ratio              # gyromagnetic ratio in (MHz/T)
        double *isotropic_chemical_shift_in_ppm # Isotropic chemical shift (Hz)
        double *shielding_symmetric_zeta_in_ppm # Nuclear shielding anisotropy (Hz)
        double *shielding_symmetric_eta         # Nuclear shielding asymmetry
        double *shielding_orientation           # Nuclear shielding PAS to CRS euler angles (rad.)
        double *quadrupolar_Cq_in_Hz            # Quadrupolar coupling constant (Hz)
        double *quadrupolar_eta                 # Quadrupolar asymmetry parameter
        double *quadrupolar_orientation         # Quadrupolar PAS to CRS euler angles (rad.)

    ctypedef struct coupling_struct:
        int number_of_couplings            # Number of couplings
        int *site_index                    # The site indexes of the coupled sites.
        double *isotropic_j_in_Hz          # isotropic J-coupling (Hz).
        double *j_symmetric_zeta_in_Hz     # J-coupling anisotropy (Hz).
        double *j_symmetric_eta            # J-coupling asymmetry.
        double *j_orientation              # J tensor PAS to CRS euler angles (rad.)
        double *dipolar_coupling_in_Hz     # Dipolar coupling constant (Hz)
        double *dipolar_eta                # Dipolar asymmetry parameter
        double *dipolar_orientation        # Dipolar tensor PAS to CRS euler angles (rad.)


cdef extern from "method.h":
    ctypedef struct MRS_event:
        double fraction                    # The weighted frequency contribution from the event.
        double magnetic_flux_density_in_T  #  he magnetic flux density in T.
        double rotor_angle_in_rad          # The rotor angle in radians.
        double rotor_frequency_in_Hz       # The sample rotation frequency in Hz.

    ctypedef struct MRS_dimension:
        int count                       #  The number of coordinates along the dimension.
        double increment                # Increment of coordinates along the dimension.
        double coordinates_offset       #  Start coordinate of the dimension.
        MRS_event *events               # Holds a list of events.
        unsigned int n_events           # The number of events.

    MRS_dimension *MRS_create_dimensions(
        MRS_averaging_scheme *scheme,
        int *count,
        double *coordinates_offset,
        double *increment,
        double *fraction,
        double *magnetic_flux_density_in_T,
        double *rotor_frequency_in_Hz,
        double *rotor_angle_in_rad,
        int *n_events,
        unsigned int n_dim,
        unsigned int number_of_sidebands)

    void MRS_free_dimension(MRS_dimension *dimensions, int n)


cdef extern from "simulation.h":
    void mrsimulator_core(
        # spectrum information and related amplitude
        double * spec,
        double spectral_start,
        double spectral_increment,
        int number_of_points,

        site_struct *sites,
        coupling_struct *couplings,

        MRS_dimension *dimensions[],
        int n_dimension,

        int quad_second_order,                    # Quad theory for second order,

        # spin rate, spin angle and number spinning sidebands
        unsigned int number_of_sidebands,
        double rotor_frequency_in_Hz,
        double rotor_angle_in_rad,

        float *transition_pathway, # Pointer to a list of transitions.
        int integration_density,
        unsigned int integration_volume,  # 0-octant, 1-hemisphere, 2-sphere
        bool_t interpolation,
        bool_t *freq_contrib,
        double *affine_matrix,
        )

    void __mrsimulator_core(
        # spectrum information and related amplitude
        double * spec,
        site_struct *sites,
        coupling_struct *couplings,
        float *transition_pathway,    # Pointer to a list of transitions.
        int n_dimension,              # the number of dimensions.
        MRS_dimension *dimensions,    # the dimensions within method.
        MRS_fftw_scheme *fftw_scheme, # the fftw scheme
        MRS_averaging_scheme *scheme, # the powder averaging scheme
        bool_t interpolation,
        bool_t *freq_contrib,
        double *affine_matrix,
        )



# -*- coding: utf-8 -*-
from .czjzek import CzjzekDistribution  # noqa: F401
from .czjzek import ExtCzjzekDistribution  # noqa: F401

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"



# -*- coding: utf-8 -*-
import warnings
from copy import deepcopy
from typing import ClassVar
from typing import Dict
from typing import List

import csdmpy as cp
import numpy as np
from mrsimulator.utils.parseable import Parseable
from pydantic import Field

from .event import Event

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class SpectralDimension(Parseable):
    r"""Base SpectralDimension class defines a spectroscopic dimension of the method.

    Attributes
    ----------

    count: int (optional).
        The number of points, :math:`N`, along the spectroscopic dimension. The default
        value is 1024.

    spectral_width: float (optional).
        The spectral width, :math:`\Delta x`, of the spectroscopic dimension in units
        of Hz. The default value is 25000.

    reference_offset: float (optional).
        The reference offset, :math:`x_0`, of the spectroscopic dimension in units of
        Hz. The default value is 0.

    origin_offset: float (optional).
        The origin offset (Larmor frequency) along the spectroscopic dimension in units
        of Hz. The default value is None. When the value is None, the origin offset is
        set to the Larmor frequency of the isotope from the
        :attr:`~mrsimulator.Method.channels` attribute of the method.

    label: str (optional).
        The value is a label of the spectroscopic dimension. The default value is None.

    description: str (optional).
        The value is a description of the spectroscopic dimension. The default value is
        None.

    events: A list of :ref:`event_api` or equivalent dict objects (optional).
        The value describes a series of events along the spectroscopic dimension.
    """

    count: int = Field(1024, gt=0)
    spectral_width: float = Field(default=25000.0, gt=0)
    reference_offset: float = Field(default=0.0)
    origin_offset: float = None
    label: str = None
    description: str = None
    events: List[Event] = []

    property_unit_types: ClassVar = {
        "spectral_width": ["frequency", "dimensionless"],
        "reference_offset": ["frequency", "dimensionless"],
        "origin_offset": ["frequency", "dimensionless"],
    }

    property_default_units: ClassVar = {
        "spectral_width": ["Hz", "ppm"],
        "reference_offset": ["Hz", "ppm"],
        "origin_offset": ["Hz", "ppm"],
    }

    property_units: Dict = {
        "spectral_width": "Hz",
        "reference_offset": "Hz",
        "origin_offset": "Hz",
    }

    class Config:
        validate_assignment = True

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """
        Parse the physical quantities of a SpectralDimension object from a
        python dictionary object.

        Args:
            dict py_dict: Dict object
        """
        py_dict_copy = deepcopy(py_dict)
        if "events" in py_dict_copy:
            py_dict_copy["events"] = [
                Event.parse_dict_with_units(e) for e in py_dict_copy["events"]
            ]

        return super().parse_dict_with_units(py_dict_copy)

    def coordinates_Hz(self) -> np.ndarray:
        r"""
        The grid coordinates along the dimension in units of Hz, evaluated as

        .. math::
            x_\text{Hz} = \left([0, 1, ... N-1] - T\right) \frac{\Delta x}{N} + x_0

        where :math:`T=N/2` and :math:`T=(N-1)/2` for even and odd values of
        :math:`N`, respectively.
        """
        n = self.count
        Tk = int(n / 2)
        increment = self.spectral_width / self.count
        return (np.arange(n) - Tk) * increment + self.reference_offset

    def coordinates_ppm(self) -> np.ndarray:
        r"""
        The grid coordinates along the dimension as dimension frequency ratio
        in units of ppm. The coordinates are evaluated as

        .. math::
            x_\text{ppm} = \frac{x_\text{Hz}} {x_0 + \omega_0}

        where :math:`\omega_0` is the Larmor frequency.
        """
        if self.origin_offset is None:
            warnings.warn(
                UserWarning(
                    "The coordinates along the dimension without an origin offset "
                    "cannot be converted to dimensionless frequency ratio."
                )
            )
        else:
            denominator = (self.origin_offset - self.reference_offset) / 1e6
            return self.coordinates_Hz() / abs(denominator)

    def to_csdm_dimension(self) -> cp.Dimension:
        """Return the spectral dimension as a CSDM dimension object."""
        increment = self.spectral_width / self.count
        label = "" if self.label is None else self.label
        description = "" if self.description is None else self.description
        dim = cp.Dimension(
            type="linear",
            count=self.count,
            increment=f"{increment} Hz",
            coordinates_offset=f"{self.reference_offset} Hz",
            label=label,
            description=description,
            complex_fft=True,
            reciprocal={"coordinates_offset": f"{-1/(2*increment)} s"},
        )
        if self.origin_offset is not None:
            dim.origin_offset = f"{self.origin_offset} Hz"
        return dim



# -*- coding: utf-8 -*-
import csdmpy as cp
import numpy as np
from mrsimulator import signal_processing as sp

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


# Creating a test CSDM object.
test_data = np.zeros((40, 40))
test_data[20, :] = 1
csdm_object = cp.CSDM(
    dependent_variables=[cp.as_dependent_variable(test_data)],
    dimensions=[
        cp.LinearDimension(count=40, increment="1 s", label="0", complex_fft=True),
        cp.Dimension(
            type="linear", count=40, increment="1 K", label="1", complex_fft=True
        ),
    ],
)

csdm_object2 = csdm_object.copy()


def test_shear_01():
    processor = sp.SignalProcessor(
        operations=[
            sp.IFFT(dim_index=1),
            sp.affine.Shear(factor="-1 K/s", dim_index=1, parallel=0),
            sp.FFT(dim_index=1),
        ]
    )

    shear_data = processor.apply_operations(data=csdm_object)
    index = np.where(shear_data.y[0].components[0] > 0.99999999)

    a = np.arange(40)
    assert np.allclose(index, [a, a])

    # complex_fft dim=0 to false
    csdm_object.x[0].complex_fft = False
    csdm_object.x[1].complex_fft = True
    shear_data = processor.apply_operations(data=csdm_object)
    index = np.where(shear_data.y[0].components[0] > 0.99999999)

    a1 = np.arange(20)
    b1 = a1 + 20
    b = np.append(b1, a1)
    assert np.allclose(index, [a, b])

    # complex_fft dim=1 to false
    csdm_object.x[0].complex_fft = True
    csdm_object.x[1].complex_fft = False
    shear_data = processor.apply_operations(data=csdm_object)
    index = np.where(shear_data.y[0].components[0] > 0.99999999)

    b = np.arange(40)
    b[1:] = a[::-1][:-1]
    assert np.allclose(index, [a, b])

    # both complex_fft set to false
    csdm_object.x[0].complex_fft = False
    csdm_object.x[1].complex_fft = False
    shear_data = processor.apply_operations(data=csdm_object)
    index = np.where(shear_data.y[0].components[0] > 0.99999999)

    a1 = np.arange(21)[::-1]
    b1 = a1[1:-1] + 20
    b = np.append(a1, b1)
    assert np.allclose(index, [a, b])


def test_serialization_and_parse():
    processor = sp.SignalProcessor(
        operations=[
            sp.IFFT(dim_index=1),
            sp.affine.Shear(factor="-1 K/s", dim_index=1, parallel=0),
            sp.FFT(dim_index=1),
        ]
    )
    serialize = processor.json()

    expected = {
        "operations": [
            {"dim_index": 1, "function": "IFFT"},
            {
                "dim_index": 1,
                "factor": "-1.0 K / s",
                "function": "affine",
                "parallel": 0,
                "type": "Shear",
            },
            {"dim_index": 1, "function": "FFT"},
        ],
    }

    assert serialize == expected

    recovered = sp.SignalProcessor.parse_dict_with_units(serialize)
    assert recovered == processor


def test_scale():
    c_inc = csdm_object.x[1].increment.value
    c_off = csdm_object.x[1].coordinates_offset.value

    processor = sp.SignalProcessor(operations=[sp.affine.Scale(factor=2, dim_index=1)])
    scaled_data = processor.apply_operations(data=csdm_object)

    s_inc = scaled_data.x[1].increment.value
    s_off = scaled_data.x[1].coordinates_offset.value

    assert s_inc == 2 * c_inc
    assert s_off == 2 * c_off
